"""
<Program>
  lockserver_daemon.py

<Started>
  30 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is the XML-RPC Backend that is used by various components of
  SeattleGeni.
 

XML-RPC Interface:
 
 TODO: describe the interface
"""

import datetime
import sys
import time
import traceback

import thread

# These are used to build a single-threaded XMLRPC server.
import SocketServer
import SimpleXMLRPCServer

import xmlrpclib

# To send the admins emails when there's an unhandled exception.
import django.core.mail 

# We use django.db.reset_queries() to prevent memory "leaks" due to query
# logging when settings.DEBUG is True.
import django.db 

# The config module contains the authcode that is required for performing
# privileged operations.
import seattlegeni.backend.config

from seattlegeni.common.api import keydb
from seattlegeni.common.api import keygen
# The lockserver is needed by the vessel cleanup thread.
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import maindb
from seattlegeni.common.api import nodemanager

from seattlegeni.common.exceptions import *

from seattlegeni.common.util import log
from seattlegeni.common.util import parallel

from seattlegeni.common.util.assertions import *

from seattlegeni.common.util.decorators import log_function_call
from seattlegeni.common.util.decorators import log_function_call_without_first_argument

from seattlegeni.website import settings


# The port that we'll listen on.
LISTENPORT = 8020





class ThreadedXMLRPCServer(SocketServer.ThreadingMixIn, SimpleXMLRPCServer.SimpleXMLRPCServer):
  """This is a threaded XMLRPC Server. """
  




def _get_node_handle_from_nodeid(nodeid, owner_pubkey=None):
  
  # Raises DoesNotExistError if no such node exists.
  node = maindb.get_node(nodeid)
  
  # If a specific owner key wasn't provided, use the one that is set in the
  # database for this node.
  if owner_pubkey is None:
    owner_pubkey = node.owner_pubkey
    
  # Raises DoesNotExistError if no such key exists.
  owner_privkey = keydb.get_private_key(owner_pubkey)
  
  return nodemanager.get_node_handle(nodeid, node.last_known_ip, node.last_known_port, owner_pubkey, owner_privkey)

    



def _assert_number_of_arguments(functionname, args, exact_number):
  """
  <Purpose>
    Ensure that an args tuple which one of the public xmlrpc functions was
    called with has an expected number of arguments.
  <Arguments>
    functionname:
      The name of the function whose number of arguments are being checked.
      This is just for logging in the case that the arguments don't match.
    args:
      A tuple of arguments (received by the other function through *args).
    exact_number:
      The exact number of arguments that must be in the args tuple.
  <Exceptions>
    Raises InvalidRequestError if args does not contain exact_number
    items.
  <Side Effects>
    None.
  <Returns>
    None.
  """
  if len(args) != exact_number:
    message = "Invalid number of arguments to function " + functionname + ". "
    message += "Expected " + str(exact_number) + ", received " + str(len(args)) + "."
    raise InvalidRequestError(message)





def _assert_valid_authcode(authcode):
  if authcode != seattlegeni.backend.config.authcode:
    raise InvalidRequestError("The provided authcode (" + authcode + ") is invalid.")





class BackendPublicFunctions(object):
  """
  All public functions of this class are automatically exposed as part of the
  xmlrpc interface.
  """
  
  def _dispatch(self, method, args):
    """
    We provide a _dispatch function (which SimpleXMLRPCServer looks for and
    uses) so that we can log exceptions due to our programming errors within
    the backend as well to detect incorrect usage by clients.
    """
      
    try:
      # Get the requested function (making sure it exists).
      try:
        func = getattr(self, method)
      except AttributeError:
        raise InvalidRequestError("The requested method '" + method + "' doesn't exist.")
      
      # Call the requested function.
      return func(*args)
    
    except NodemanagerCommunicationError, e:
      raise xmlrpclib.Fault(100, "Node communication failure: " + str(e))
    
    except (DoesNotExistError, InvalidRequestError, AssertionError):
      log.error("The backend was used incorrectly: " + traceback.format_exc())
      raise
    
    except:
      # We assume all other exceptions are bugs in the backend. Unlike the
      # lockserver where it might result in broader data corruption, here in
      # the backend we allow the backend to continue serving other requests.
      # That is, we don't go through steps to try to shutdown the backend.
      
      message = "The backend had an internal error: " + traceback.format_exc()
      log.critical(message)
      
      # Send an email to the addresses listed in settings.ADMINS
      if not settings.DEBUG:
        subject = "Critical SeattleGeni backend error"
        django.core.mail.mail_admins(subject, message)
      
      raise
      




  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  @log_function_call
  def GenerateKey(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('GenerateKey', args, 1)
    
    keydescription = args[0]
    
    assert_str(keydescription)
    
    # Generate a new keypair.
    (pubkey, privkey) = keygen.generate_keypair()
    
    # Store the private key in the keydb.
    keydb.set_private_key(pubkey, privkey, keydescription)
    
    # Return the public key as a string.
    return pubkey





  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  @log_function_call
  def SetVesselUsers(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('SetVesselUsers', args, 3)
    (nodeid, vesselname, userkeylist) = args
    
    assert_str(nodeid)
    assert_str(vesselname)
    assert_list(userkeylist)
    
    for userkey in userkeylist:
      assert_str(userkey)

    # Note: The nodemanager checks whether each key is a valid key and will
    #       raise an exception if it is not.
      
    # Raises a DoesNotExistError if there is no node with this nodeid.
    nodehandle = _get_node_handle_from_nodeid(nodeid)
    
    # Raises NodemanagerCommunicationError if it fails.
    nodemanager.change_users(nodehandle, vesselname, userkeylist)





  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  @log_function_call_without_first_argument
  def SetVesselOwner(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('SetVesselOwner', args, 5)
    (authcode, nodeid, vesselname, old_ownerkey, new_ownerkey) = args
    
    assert_str(authcode)
    assert_str(nodeid)
    assert_str(vesselname)
    assert_str(old_ownerkey)
    assert_str(new_ownerkey)
    
    _assert_valid_authcode(authcode)
    
    # Note: The nodemanager checks whether the owner key is a valid key and
    #       will raise an exception if it is not.
    
    # Raises a DoesNotExistError if there is no node with this nodeid.
    nodehandle = _get_node_handle_from_nodeid(nodeid, owner_pubkey=old_ownerkey)
    
    # Raises NodemanagerCommunicationError if it fails.
    nodemanager.change_owner(nodehandle, vesselname, new_ownerkey)
    
    
    
 
    
  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  @log_function_call_without_first_argument
  def SplitVessel(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('SplitVessels', args, 4)
    (authcode, nodeid, vesselname, desiredresourcedata) = args
    
    assert_str(authcode)
    assert_str(nodeid)
    assert_str(vesselname)
    assert_str(desiredresourcedata)
    
    _assert_valid_authcode(authcode)
    
    # Raises a DoesNotExistError if there is no node with this nodeid.
    nodehandle = _get_node_handle_from_nodeid(nodeid)
    
    # Raises NodemanagerCommunicationError if it fails.
    return nodemanager.split_vessel(nodehandle, vesselname, desiredresourcedata)
    
    



  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  @log_function_call_without_first_argument
  def JoinVessels(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('JoinVessels', args, 4)
    (authcode, nodeid, firstvesselname, secondvesselname) = args
    
    assert_str(authcode)
    assert_str(nodeid)
    assert_str(firstvesselname)
    assert_str(secondvesselname)
    
    _assert_valid_authcode(authcode)
    
    # Raises a DoesNotExistError if there is no node with this nodeid.
    nodehandle = _get_node_handle_from_nodeid(nodeid)
    
    # Raises NodemanagerCommunicationError if it fails.
    return nodemanager.join_vessels(nodehandle, firstvesselname, secondvesselname)
      




def cleanup_vessels():
  """
  This function is started as separate thread. It continually checks whether
  there are vessels needing to be cleaned up and initiates cleanup as needed.
  """
  
  log.info("[cleanup_vessels] cleanup thread started.")

  # Start a transaction management.
  django.db.transaction.enter_transaction_management()

  # Run forever.
  while True:
    
    try:
      
      # Sleep a few seconds for those times where we don't have any vessels to clean up.
      time.sleep(5)
      
      # We shouldn't be running the backend in production with
      # settings.DEBUG = True. Just in case, though, tell django to reset its
      # list of saved queries each time through the loop. Note that this is not
      # specific to the cleanup thread as other parts of the backend are using
      # the maindb, as well, so we're overloading the purpose of the cleanup
      # thread by doing this here. This is just a convenient place to do it.
      # See http://docs.djangoproject.com/en/dev/faq/models/#why-is-django-leaking-memory
      # for more info.
      if settings.DEBUG:
        django.db.reset_queries()
      
      # First, make it so that expired vessels are seen as dirty. We aren't
      # holding a lock on the nodes when we do this. It's possible that we do
      # this while someone else has a lock on the node. What would result?
      # I believe the worst result is that a user has their vessel marked as
      # dirty after they renewed in the case where they are renewing it just
      # as it expires (with some exceptionally bad timing involved). And, 
      # that's not really very bad as if the user is trying to renew at the
      # exact moment it expires, their trying their luck with how fast their
      # request gets processed, anyways. In short, I don't think it's important
      # enough to either obtain locks to do this or to rewrite the code to
      # avoid any need for separately marking expired vessels as dirty rather
      # than just trying to process expired vessels directly in the code below.
      date_started=datetime.datetime.now()
      expired_list = maindb.mark_expired_vessels_as_dirty()
      if len(expired_list) > 0:
        log.info("[cleanup_vessels] " + str(len(expired_list)) + 
                 " expired vessels have been marked as dirty: " + str(expired_list))
        maindb.create_action_log_event("mark_expired_vessels_as_dirty", user=None, second_arg=None,
                                       third_arg=None, was_successful=True, message=None,
                                       date_started=date_started, vessel_list=expired_list)

      # Get a list of vessels to clean up. This doesn't include nodes known to
      # be inactive as we would just continue failing to communicate with nodes
      # that are down.
      cleanupvessellist = maindb.get_vessels_needing_cleanup()
      if len(cleanupvessellist) == 0:
        continue
        
      log.info("[cleanup_vessels] " + str(len(cleanupvessellist)) + " vessels to clean up: " + str(cleanupvessellist))
      
      parallel_results = parallel.run_parallelized(cleanupvessellist, _cleanup_single_vessel)
        
      if len(parallel_results["exception"]) > 0:
        for vessel, exception_message in parallel_results["exception"]:
          log_message = "Unhandled exception during parallelized vessel cleanup: " + exception_message
          log.critical(log_message)
        # Raise the last exceptions so that the admin gets an email.
        raise InternalError(log_message)  
    
    except:
      message = "[cleanup_vessels] Something very bad happened: " + traceback.format_exc()
      log.critical(message)
      
      # Send an email to the addresses listed in settings.ADMINS
      if not settings.DEBUG:
        subject = "Critical SeattleGeni backend error"
        django.core.mail.mail_admins(subject, message)
        
        # Sleep for ten minutes to make sure we don't flood the admins with error
        # report emails.
        time.sleep(600)
    finally:
      # Manually commit the transaction to prevent caching.
      django.db.transaction.commit()
      




def _cleanup_single_vessel(vessel):
  """
  This function is passed by cleanup_vessels() as the function argument to
  run_parallelized().
  """
  
  # This does seem wasteful of lockserver communication to require four
  # round-trips with the lockserver (get handle, lock, unlock, release handle),
  # but if we really want to fix that then I think the best thing to do would
  # be to allow obtaining a lockhandle and releasing a lockhandle to be done
  # in the same calls as lock acquisition and release. 
  
  node_id = maindb.get_node_identifier_from_vessel(vessel)
  lockserver_handle = lockserver.create_lockserver_handle()
  
  # Lock the node that the vessels is on.
  lockserver.lock_node(lockserver_handle, node_id)
  try:
    # Get a new vessel object from the db in case it was modified in the db
    # before the lock was obtained.
    vessel = maindb.get_vessel(node_id, vessel.name)
    
    # Now that we have a lock on the node that this vessel is on, find out
    # if we should still clean up this vessel (e.g. maybe a node state
    # transition script moved the node to a new state and this vessel was
    # removed).
    needscleanup, reasonwhynot = maindb.does_vessel_need_cleanup(vessel)
    if not needscleanup:
      log.info("[_cleanup_single_vessel] Vessel " + str(vessel) + 
               " no longer needs cleanup: " + reasonwhynot)
      return
    
    nodeid = maindb.get_node_identifier_from_vessel(vessel)
    nodehandle = _get_node_handle_from_nodeid(nodeid)
    
    try:
      log.info("[_cleanup_single_vessel] About to ChangeUsers on vessel " + str(vessel))
      nodemanager.change_users(nodehandle, vessel.name, [''])
      log.info("[_cleanup_single_vessel] About to ResetVessel on vessel " + str(vessel))
      nodemanager.reset_vessel(nodehandle, vessel.name)
    except NodemanagerCommunicationError:
      # We don't pass this exception up. Maybe the node is offline now. At some
      # point, it will be marked in the database as offline (should we be doing
      # that here?). At that time, the dirty vessels on that node will not be
      # in the cleanup list anymore.
      log.info("[_cleanup_single_vessel] Failed to cleanup vessel " + 
               str(vessel) + ". " + traceback.format_exc())
      return
      
    # We only mark it as clean if no exception was raised when trying to
    # perform the above nodemanager operations.
    maindb.mark_vessel_as_clean(vessel)
  
    log.info("[_cleanup_single_vessel] Successfully cleaned up vessel " + str(vessel))

  finally:
    # Unlock the node.
    lockserver.unlock_node(lockserver_handle, node_id)
    lockserver.destroy_lockserver_handle(lockserver_handle)





def sync_user_keys_of_vessels():
  """
  This function is started as separate thread. It continually checks whether
  there are vessels needing their user keys sync'd and initiates the user key
  sync as needed.
  """

  log.info("[sync_user_keys_of_vessels] thread started.")

  # Run forever.
  while True:
    
    try:
      
      # Sleep a few seconds for those times where we don't have any vessels to clean up.
      time.sleep(5)
      
      # We shouldn't be running the backend in production with
      # settings.DEBUG = True. Just in case, though, tell django to reset its
      # list of saved queries each time through the loop.
      if settings.DEBUG:
        django.db.reset_queries()
      
      # Get a list of vessels that need to have user keys sync'd. This doesn't
      # include nodes known to be inactive as we would just continue failing to
      # communicate with nodes that are down.
      vessellist = maindb.get_vessels_needing_user_key_sync()
      if len(vessellist) == 0:
        continue
        
      log.info("[sync_user_keys_of_vessels] " + str(len(vessellist)) + 
               " vessels to have user keys sync'd: " + str(vessellist))
     
      parallel_results = parallel.run_parallelized(vessellist, _sync_user_keys_of_single_vessel)
     
      if len(parallel_results["exception"]) > 0:
        for vessel, exception_message in parallel_results["exception"]:
          log_message = "Unhandled exception during parallelized vessel user key sync: " + exception_message
          log.critical(log_message)
        # Raise the last exceptions so that the admin gets an email.
        raise InternalError(log_message)
        
    except:
      message = "[sync_user_keys_of_vessels] Something very bad happened: " + traceback.format_exc()
      log.critical(message)
      
      # Send an email to the addresses listed in settings.ADMINS
      if not settings.DEBUG:
        subject = "Critical SeattleGeni backend error"
        django.core.mail.mail_admins(subject, message)
        
        # Sleep for ten minutes to make sure we don't flood the admins with error
        # report emails.
        time.sleep(600)





def _sync_user_keys_of_single_vessel(vessel):
  """
  This function is passed by sync_user_keys_of_vessels() as the function
  argument to run_parallelized().
  """
  
  # This does seem wasteful of lockserver communication to require four
  # round-trips with the lockserver (get handle, lock, unlock, release handle),
  # but if we really want to fix that then I think the best thing to do would
  # be to allow obtaining a lockhandle and releasing a lockhandle to be done
  # in the same calls as lock acquisition and release. 
  
  node_id = maindb.get_node_identifier_from_vessel(vessel)
  lockserver_handle = lockserver.create_lockserver_handle()
  
  # Lock the node that the vessels is on.
  lockserver.lock_node(lockserver_handle, node_id)
  try:
    # Get a new vessel object from the db in case it was modified in the db
    # before the lock was obtained.
    vessel = maindb.get_vessel(node_id, vessel.name)
  
    # Now that we have a lock on the node that this vessel is on, find out
    # if we should still sync user keys on this vessel (e.g. maybe a node state
    # transition script moved the node to a new state and this vessel was
    # removed).
    needssync, reasonwhynot = maindb.does_vessel_need_user_key_sync(vessel)
    if not needssync:
      log.info("[_sync_user_keys_of_single_vessel] Vessel " + str(vessel) + 
               " no longer needs user key sync: " + reasonwhynot)
      return
    
    nodeid = maindb.get_node_identifier_from_vessel(vessel)
    nodehandle = _get_node_handle_from_nodeid(nodeid)
    
    # The list returned from get_users_with_access_to_vessel includes the key of
    # the user who has acquired the vessel along with any other users they have
    # given access to.
    user_list = maindb.get_users_with_access_to_vessel(vessel)
    
    key_list = []
    for user in user_list:
      key_list.append(user.user_pubkey)
      
    if len(key_list) == 0:
      raise InternalError("InternalError: Empty user key list for vessel " + str(vessel))
    
    try:
      log.info("[_sync_user_keys_of_single_vessel] About to ChangeUsers on vessel " + str(vessel))
      nodemanager.change_users(nodehandle, vessel.name, key_list)
    except NodemanagerCommunicationError:
      # We don't pass this exception up. Maybe the node is offline now. At some
      # point, it will be marked in the database as offline and won't show up in
      # our list of vessels to sync user keys of anymore.
      log.info("[_sync_user_keys_of_single_vessel] Failed to sync user keys of vessel " + 
               str(vessel) + ". " + traceback.format_exc())
      return
      
    # We only mark it as sync'd if no exception was raised when trying to perform
    # the above nodemanager operations.
    maindb.mark_vessel_as_not_needing_user_key_sync(vessel)
  
    log.info("[_sync_user_keys_of_single_vessel] Successfully sync'd user keys of vessel " + str(vessel))

  finally:
    # Unlock the node.
    lockserver.unlock_node(lockserver_handle, node_id)
    lockserver.destroy_lockserver_handle(lockserver_handle)





def main():
  
  # Initialize the main database.
  maindb.init_maindb()

  # Initialize the key database.
  keydb.init_keydb()
  
  # Initialize the nodemanager.
  nodemanager.init_nodemanager()

  # Start the background thread that does vessel cleanup.
  thread.start_new_thread(cleanup_vessels, ())
  
  # Start the background thread that does vessel user key synchronization.
  thread.start_new_thread(sync_user_keys_of_vessels, ())
  
  # Register the XMLRPCServer. Use allow_none to allow allow the python None value.
  server = ThreadedXMLRPCServer(("127.0.0.1", LISTENPORT), allow_none=True)

  log.info("Backend listening on port " + str(LISTENPORT) + ".")

  server.register_instance(BackendPublicFunctions()) 
  while True:
    server.handle_request()





if __name__ == '__main__':
  try:
    main()
  except KeyboardInterrupt:
    log.info("Exiting on KeyboardInterrupt.")
    sys.exit(0)



"""
<Program>
  test_urls.py

<Started>
  Oct 13, 2009

<Author>
  Justin Samuel

<Purpose>
  
<Notes>
  Uses a Django test client, not our xmlrpc client. This is intentional.
  The xmlrpc client should be tested separately.
"""

#pragma out
#pragma error OK


# We import the testlib FIRST, as the test db settings 
# need to be set before we import anything else.
from seattlegeni.tests import testlib

import xmlrpclib
import unittest

from django.contrib.auth.models import User as DjangoUser

from seattlegeni.common.exceptions import *

from seattlegeni.common.api import maindb

from seattlegeni.website.xmlrpc import views

from seattlegeni.website.xmlrpc.tests import xmlrpctestutil

from seattlegeni.website.control import interface
from seattlegeni.website.control import models

from seattle.repyportability import *
add_dy_support(locals())

dy_import_module_symbols("rsa.r2py")




def mock_raises_DoesNotExistError(*args, **kwargs):
  raise DoesNotExistError



def mock_raises_InvalidRequestError(*args, **kwargs):
  raise InvalidRequestError



def mock_raises_InsufficientUserResourcesError(*args, **kwargs):
  raise InsufficientUserResourcesError



def mock_raises_UnableToAcquireResourcesError(*args, **kwargs):
  raise UnableToAcquireResourcesError



def mock_noop(*args, **kwargs):
  pass



def mock_interface_get_user_with_api_key(username, api_key):
  geniuser = models.GeniUser(username=username, password="password", email='test@test.com',
                             affiliation='test affil', user_pubkey='user_pubkey',
                             user_privkey='user_privkey', donor_pubkey='donor_pubkey',
                             usable_vessel_port='12345', free_vessel_credits=10)
  geniuser.api_key = api_key
  # Not saving the geniuser record, we're not trying to interact with the db.
  return geniuser



def mock_interface_get_user_with_password(username, password):
  geniuser = models.GeniUser(username=username, password=password, email='test@test.com',
                             affiliation='test affil', user_pubkey='user_pubkey',
                             user_privkey='user_privkey', donor_pubkey='donor_pubkey',
                             usable_vessel_port='12345', free_vessel_credits=10)
  geniuser.password = password
  # Not saving the geniuser record, we're not trying to interact with the db.
  return geniuser



def create_mock_get_user(username, password="testpassword", api_key="testapikey",
                         pubkeystr="pubkeystr", privkeystr="privkeystr"):
  
  geniuser = models.GeniUser(username=username, password=password, email='test@test.com',
                             affiliation='test affil', user_pubkey=pubkeystr,
                             user_privkey=privkeystr, donor_pubkey='donor_pubkey',
                             usable_vessel_port='12345', free_vessel_credits=10,
                             api_key = api_key)
  # Not saving the geniuser record, we're not trying to interact with the db.

  def mock_interface_get_user(username, *args, **kwargs):
    return geniuser

  return mock_interface_get_user



def mock_interface_get_vessel_list(vesselhandle_list):
  return []



def mock_interface_get_vessel_infodict_list(vessel_list):
  return []


def mock_interface_regenerate_api_key(geniuser):
  return "a" * maindb.API_KEY_LENGTH


def mock_interface_change_user_keys(geniuser, pubkeystring):
  pass


proxy = xmlrpclib.ServerProxy('http://fakehost/xmlrpc/',
                              transport=xmlrpctestutil.TestTransport())




class SeattleGeniTestCase(unittest.TestCase):


  def setUp(self):
    # Indicate a user record was found with the provided auth info.
    interface.get_user_with_api_key = mock_interface_get_user_with_api_key
    interface.get_user_with_password = mock_interface_get_user_with_password


  def tearDown(self):
    pass



  def test_invalid_auth_dict(self):
    
    auth = 'not a dict'
    
    try:
      proxy.get_account_info(auth)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    auth = {'username':'tester'}
    
    try:
      proxy.get_account_info(auth)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    
    
  def test_all_functions_auth_failure(self):
    """
    Make sure all of the xmlrpc calls raise an exception if the auth check
    fails.
    """
    
    auth = {'username':'tester', 'api_key':'api_key'}

    # Indicate no user record was found with the provided auth info.
    interface.get_user_with_api_key = mock_raises_DoesNotExistError
    
    try:
      proxy.renew_resources(auth, [])
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")
    
    try:
      proxy.acquire_resources(auth, {})
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")
      
    try:
      proxy.acquire_specific_vessels(auth, [])
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")
    
    try:
      proxy.release_resources(auth, [])
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")

    try:
      proxy.get_resource_info(auth)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")

    try:
      proxy.get_account_info(auth)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")

    try:
      proxy.get_public_key(auth)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_AUTHERROR)
    else:
      self.fail("Expected an exception.")

    

  def test_renew_resources_invalid_vessel_handle_list(self):
    
    auth = {'username':'tester', 'api_key':'api_key'}

    vesselhandle_list = {'dict not a list':'123'}
    
    try:
      proxy.renew_resources(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")


  
  def test_renew_resources(self):
    
    auth = {'username':'tester', 'api_key':'api_key'}
    vesselhandle_list = ['vessel1', 'vessel2']
    
    # Check successful case. Returns a zero to indicate success.
    
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.renew_vessels = mock_noop
    
    response = proxy.renew_resources(auth, vesselhandle_list)
    self.assertEqual(response, 0)
    
    # Check various error cases. These all return FAULTCODE_INVALIDREQUEST.
    
    interface.get_vessel_list = mock_raises_DoesNotExistError
    interface.renew_vessels = mock_noop
    
    try:
      proxy.renew_resources(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    interface.get_vessel_list = mock_raises_InvalidRequestError
    interface.renew_vessels = mock_noop
    
    try:
      proxy.renew_resources(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
  
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.renew_vessels = mock_raises_InvalidRequestError
    
    try:
      proxy.renew_resources(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    # This is the case where the user has less credits than acquired resources.
    
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.renew_vessels = mock_raises_InsufficientUserResourcesError
    
    try:
      proxy.renew_resources(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_NOTENOUGHCREDITS)
    else:
      self.fail("Expected an exception.")


  
  def test_acquire_resources_rspecs(self):
    
    auth = {'username':'tester', 'api_key':'api_key'}
    vesselhandle_list = ['vessel1', 'vessel2']
    
    rspec_valid = {'rspec_type':'random', 'number_of_nodes':2}
    
    # invalid rspecs
    rspec_negativenodes = {'rspec_type':'random', 'number_of_nodes':-10}
    rspec_invalid_rtype = {'rspec_type':'wtf', 'number_of_nodes':10}
    rspec_badtype = {'rspec_type':42, 'number_of_nodes':'a string'}
    rspec_empty = {'rspec_type':'', 'number_of_nodes':0}
    rspec_missingkey = {'number_of_nodes':10}
    rspec_badkeys = {'bad':'', 'keys':''}
    
    try:
      proxy.acquire_resources(auth, rspec_negativenodes)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
      
    try:
      proxy.acquire_resources(auth, rspec_invalid_rtype)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    try:
      proxy.acquire_resources(auth, rspec_badtype)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
      
    try:
      proxy.acquire_resources(auth, rspec_empty)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    try:
      proxy.acquire_resources(auth, rspec_missingkey)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
    
    try:
      proxy.acquire_resources(auth, rspec_badkeys)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")
      
    # This is the case where the user has less credits than the number of
    # vessels they are trying to acquire.
    
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.acquire_vessels = mock_raises_InsufficientUserResourcesError
    
    try:
      proxy.acquire_resources(auth, rspec_valid)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_NOTENOUGHCREDITS)
    else:
      self.fail("Expected an exception.")
      
    # This is the case where SeattleGENI doesn't have enough resources to
    # acquire the requested vessels.

    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.acquire_vessels = mock_raises_UnableToAcquireResourcesError
    
    try:
      proxy.acquire_resources(auth, rspec_valid)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_UNABLETOACQUIRE)
    else:
      self.fail("Expected an exception.")



  def test_acquire_specific_vessels(self):
    
    auth = {'username':'tester', 'api_key':'api_key'}
    vesselhandle_list = ['vessel1', 'vessel2']
    
    # Check successful case. Returns a zero to indicate success.
    
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.get_vessel_infodict_list = mock_interface_get_vessel_infodict_list
    interface.acquire_specific_vessels = mock_noop
    
    response = proxy.acquire_specific_vessels(auth, vesselhandle_list)
    self.assertTrue(isinstance(response, list))
    
    # vesselhandle_list is not a list.
    try:
      proxy.acquire_specific_vessels(auth, 1)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_INVALIDREQUEST)
    else:
      self.fail("Expected an exception.")

    # This is the case where the user has less credits than the number of
    # vessels they are trying to acquire.
    
    interface.get_vessel_list = mock_interface_get_vessel_list
    interface.acquire_specific_vessels = mock_raises_InsufficientUserResourcesError
    
    try:
      proxy.acquire_specific_vessels(auth, vesselhandle_list)
    except xmlrpclib.Fault, e:
      self.assertEqual(e.faultCode, views.FAULTCODE_NOTENOUGHCREDITS)
    else:
      self.fail("Expected an exception.")



  def test_get_encrypted_api_key(self):
    
    username = "testuser"
    api_key = "testapikey"
    
    (pubkeydict, privkeydict) = rsa_gen_pubpriv_keys(512)
    pubkeystr = rsa_publickey_to_string(pubkeydict)
    privkeystr = rsa_privatekey_to_string(privkeydict)
    
    interface.get_user_without_password = create_mock_get_user(username, api_key=api_key,
                                                               pubkeystr=pubkeystr,
                                                               privkeystr=privkeystr)
    
    response = proxy.get_encrypted_api_key(username)
    self.assertTrue(isinstance(response, str))
    
    encrypted_data = proxy.get_encrypted_api_key(username)
    decrypted_data = rsa_decrypt(encrypted_data, privkeydict)
    split_data = decrypted_data.split("!")
    
    self.assertEqual(api_key, split_data[1])



  def test_regenerate_api_key(self):
    
    username = "testuser"
    password = "testpassword"
    original_api_key = "original_api_key"
    
    interface.get_user_without_password = create_mock_get_user(username, password=password,
                                                               api_key=original_api_key)
    interface.regenerate_api_key = mock_interface_regenerate_api_key
    
    pwauth = {"username" : username, "password" : password}
    
    response = proxy.regenerate_api_key(pwauth)
    self.assertTrue(isinstance(response, str))
    self.assertNotEqual(response, original_api_key)
    self.assertEqual(len(response), maindb.API_KEY_LENGTH)



  def test_set_public_key(self):
    
    username = "testuser"
    password = "testpassword"
    original_api_key = "original_api_key"
    
    interface.get_user_without_password = create_mock_get_user(username, password=password,
                                                               api_key=original_api_key)
    interface.change_user_keys = mock_interface_change_user_keys
    
    pwauth = {"username" : username, "password" : password}
    
    (pubkeydict, privkeydict) = rsa_gen_pubpriv_keys(512)
    pubkeystr = rsa_publickey_to_string(pubkeydict)
    
    proxy.set_public_key(pwauth, pubkeystr)



def main():
  # The tests don't use the database, so we just make a single database
  # so that it exists to prevent unrelated errors when we use the models.
  testlib.setup_test_environment()
  testlib.setup_test_db()
  
  try:
    unittest.main()
  
  finally:
    testlib.teardown_test_db()
    testlib.teardown_test_environment()



if __name__ == "__main__":
  main()



# Modules to make available for convenience so the names are available in the
# ipython shell. Just a lazy way to not have to execute these lines individually
# in a new shell.
from seattlegeni.website.control import interface
from seattlegeni.common.api import backend
from seattlegeni.common.api import keydb
from seattlegeni.common.api import keygen
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import maindb
from seattlegeni.common.api import nodemanager

# grab a few objects to play with
g = maindb.get_user('user0')
(v, v2) = maindb.get_available_wan_vessels(g, 2)[:2]




import unittest

import lockserver_daemon as lockserver


class TheTestCase(unittest.TestCase):

  def setUp(self):
    # Reset the lockserver's global variables between each test.
    lockserver.init_globals()


  def testLockdictWithMultipleLockTypes(self):
    
    sess = []
    sess.append(lockserver.do_start_session())
    
    # It is not allowed to make a single request for multiple types of locks.
    locks = {'user':['bob'], 'node':['123']}
    
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)


  def testRequestForTypeOfLockAlreadyHeld(self):
    
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    
    # First session gets a lock on user 'bob'
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[0], locks)
    
    # First session then tries to get another user lock, which it shouldn't be
    # allowed to do while it already holds a user lock.
    locks = {'user':['alice']}
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)

    # Second session gets a lock on node '123'.
    locks = {'node':['123']}
    lockserver.do_acquire_locks(sess[1], locks)
    
    # Second session then tries to get another node lock, which it shouldn't be
    # allowed to do while it already holds a node lock.
    locks = {'node':['123']}
    
    func = lockserver.do_acquire_locks
    args = (sess[1], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)

    
  def testRequestForUserLockWhileHoldingNodeLock(self):
    
    sess = []
    sess.append(lockserver.do_start_session())
    
    # First session gets a lock on node '123'
    locks = {'node':['123']}
    lockserver.do_acquire_locks(sess[0], locks)
    
    # First session then tries to get a user lock, which it shouldn't be
    # allowed to do while it already holds a node lock.
    locks = {'user':['bob']}
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)




#pragma out
#pragma error OK

# The seattlegeni testlib must be imported first.
from seattlegeni.tests import testlib

from seattlegeni.tests import mocklib

from seattlegeni.common.api import maindb

from seattlegeni.common.exceptions import *

from seattlegeni.common.util import validations

from seattlegeni.website.control import interface

import unittest



mocklib.mock_lockserver_calls()


good_username = "myusername"
good_password = "mypassword"
good_email = "example@example.com"
good_affiliation = "my affiliation"
good_pubkey = "400 401"


class SeattleGeniTestCase(unittest.TestCase):


  def setUp(self):
    # Setup a fresh database for each test.
    testlib.setup_test_db()



  def tearDown(self):
    # Cleanup the test database.
    testlib.teardown_test_db()



  def _create_user_expect_success(self, username, password=good_password, email=good_email, 
                                  affiliation=good_affiliation, pubkey=good_pubkey):
    
    # We expect a single key to be generated through the backed (the donor key)
    mocklib.mock_backend_generate_key(["1 2"])
    
    created_user = interface.register_user(username, password, email, affiliation, pubkey)
  
    user_from_db = maindb.get_user(username)
    assert(user_from_db.username == created_user.username)
    assert(user_from_db.email == created_user.email)
    assert(user_from_db.affiliation == created_user.affiliation)
    assert(user_from_db.user_pubkey == created_user.user_pubkey)
    assert(user_from_db.user_privkey == created_user.user_privkey)
    assert(user_from_db.donor_pubkey == created_user.donor_pubkey)
  
  
  
  def _create_user_expect_validation_error(self, username, password=good_password, email=good_email, 
                                  affiliation=good_affiliation, pubkey=good_pubkey):
    
    # An empty list because the ValidationError should occur before any backend
    # generate_key() calls are performed.
    mocklib.mock_backend_generate_key([])
    
    func = interface.register_user
    args = (username, password, email, affiliation, pubkey)
    self.assertRaises(ValidationError, func, *args)



  def test_username_minimum_length(self):
    username = "a" * validations.USERNAME_MIN_LENGTH
    userobj = self._create_user_expect_success(username)



  def test_username_too_short(self):
    username = "a" * (validations.USERNAME_MIN_LENGTH - 1)
    self._create_user_expect_validation_error(username)



  def test_username_maximum_length(self):
    username = "a" * validations.USERNAME_MAX_LENGTH
    userobj = self._create_user_expect_success(username)



  def test_username_too_long(self):
    username = "a" * (validations.USERNAME_MAX_LENGTH + 1)
    self._create_user_expect_validation_error(username)



  def test_username_invalid_chars_spaces(self):
    self._create_user_expect_validation_error("test user") # space in middle
    self._create_user_expect_validation_error(" testuser") # leading space
    self._create_user_expect_validation_error("testuser ") # trailing space



  def test_valid_chars_nonleading_underscores(self):
    self._create_user_expect_success("9test_user") 
    self._create_user_expect_success("test9user_") # one trailing underscore
    self._create_user_expect_success("testuser__") # two trailing underscores
    
    

  def test_username_invalid_chars_leading_underscores(self):
    username = "_testuser"
    self._create_user_expect_validation_error(username)



  def test_password_too_short(self):
    bad_password = "a" * (validations.PASSWORD_MIN_LENGTH - 1)
    self._create_user_expect_validation_error(good_username, password=bad_password)



  def test_password_same_as_username(self):
    self._create_user_expect_validation_error(good_username, password=good_username)



  def test_valid_country_email(self):
    email = "test@example.co.uk"
    self._create_user_expect_success(good_username, email=email)



  def test_valid_gmail_label_email(self):
    email = "test#label@gmail.com"
    self._create_user_expect_success(good_username, email=email)



  def test_invalid_email(self):
    bad_email = "test@test" # missing expected tld
    self._create_user_expect_validation_error(good_username, email=bad_email)



  def test_affiliation_too_short(self):
    bad_affiliation = "a" * (validations.AFFILIATION_MIN_LENGTH - 1)
    self._create_user_expect_validation_error(good_username, affiliation=bad_affiliation)



  def test_affiliation_too_long(self):
    bad_affiliation = "a" * (validations.AFFILIATION_MAX_LENGTH + 1)
    self._create_user_expect_validation_error(good_username, affiliation=bad_affiliation)



  def test_invalid_user_pubkey_empty_string(self):
    """
    Tests an empty string for the pubkey. It should be None rather than an
    empty string to indicate that we should generate the key for the user.
    """
    bad_pubkey = ""
    self._create_user_expect_validation_error(good_username, pubkey=bad_pubkey)
  
  
  
  def test_invalid_user_pubkey_invalid_key(self):
    bad_pubkey = "0" # should be two numbers
    self._create_user_expect_validation_error(good_username, pubkey=bad_pubkey)
   
    bad_pubkey = "1 0" # first number must be smaller than second
    self._create_user_expect_validation_error(good_username, pubkey=bad_pubkey)
  
    bad_pubkey = "a b" # letters, not numbers
    self._create_user_expect_validation_error(good_username, pubkey=bad_pubkey)
  
    bad_pubkey = "2 3 3" # they might have tried to upload their private key
    self._create_user_expect_validation_error(good_username, pubkey=bad_pubkey)


  
  def test_seattlegeni_generates_user_keypair(self):
    
    # We expect a single keypair to be generated directly through the keygen
    # api (specifically, the user keys).
    user_pubkey = "3 4"
    user_privkey = "2 3 3"
    mocklib.mock_keygen_generate_keypair([(user_pubkey, user_privkey)])
    
    # We expect a single key to be generated directly through the backed api
    # (specifically, the donor key).
    donor_pubkey = "1 2"
    mocklib.mock_backend_generate_key([donor_pubkey])
    
    provided_pubkey=None
    interface.register_user(good_username, good_password, good_email, 
                                           good_affiliation, provided_pubkey)

    user_from_db = maindb.get_user(good_username)

    assert(user_from_db.user_pubkey == user_pubkey)
    assert(user_from_db.user_privkey == user_privkey)
    assert(user_from_db.donor_pubkey == donor_pubkey)



# TODO: test the number of free vessel credits the user has after creation.
# TODO: test the number of used vessel credits the user has after creation.
# TODO: test the number of donations the user has after creation.
# TODO: test username exists




def run_test():
  unittest.main()



if __name__ == "__main__":
  run_test()



"""
<Program>
  parallel.py

<Started>
  14 August 2009

<Author>
  Justin Samuel

<Purpose>
  This module just provides a convenience function for using the repy
  parallelize module for parallelizing a function call. This is just to
  keep other code clean and prevent accidental leakage of the handles
  provided by the repy module for parallelized function.
"""

import time
import traceback

from seattle.repyportability import *
add_dy_support(locals())

dy_import_module_symbols("parallelize.r2py")





# The number of threads to be running at a given time for each parallelized
# function call.
CONCURRENT_THREADS_PER_CALL = 10





def run_parallelized(first_arg_list, func, *additional_args):
  """
   <Purpose>
      Call a function with each argument in a list in parallel.
   <Arguments>
      first_arg_list:
          The list of arguments the function should be called with. Each
          argument is passed once to the function.
      func:
          The function to call
      additional_args:
          Extra arguments the function should be called with (every function
          is passed the same extra args).
   <Exceptions>
      InternalError
        If running the parallelized function failed (failed to run at all, not
        just had each function call raise an exception).
   <Side Effects>
      Creates multiple threads to run func in parallel.
   <Returns>
      A dictionary with the results.   The format is
        {'exception':list of tuples with (target, exception string), 
         'aborted':list of targets, 'returned':list of tuples with (target, 
         return value)}
  """
  
  try:
    phandle = parallelize_initfunction(first_arg_list, func, CONCURRENT_THREADS_PER_CALL, *additional_args)
  
    try: 
      while not parallelize_isfunctionfinished(phandle):
        time.sleep(0.1)
  
      return parallelize_getresults(phandle)
  
    finally:
      # clean up the handle
      parallelize_closefunction(phandle)
      
  except ParallelizeError, e:
    raise InternalError("Failed to run parallelized function: " + traceback.format_exc())



#!/usr/bin/env python
"""
Populates the database with some sample data.
"""

from seattlegeni.common.api import maindb

# Decrease the amount of logging output.
from seattlegeni.common.util import log
log.loglevel = log.LOG_LEVEL_ERROR



# Delete everything from the database. To be safe, this isn't enabled by
# default. You should probably use ./managepy --flush, instead.
#Donation.objects.all().delete()
#GeniUser.objects.all().delete()
#Node.objects.all().delete()
#Vessel.objects.all().delete()
#VesselPort.objects.all().delete()
#VesselUserAccessMap.objects.all().delete()


# Create GeniUser records
user_list = []
for i in range(10):
  username = 'user' + str(i)
  user = maindb.create_user(username, 'mypassword', 'myemail', 'myaff', 'mypubkey', 'myprivkey', 'mydonorpubkey')
  user.save()
  user_list.append(user)

# Create Node records
node_list = []
for i in range(100):
  node_identifier = 'node' + str(i)
  extra_vessel_name = 'v2'
  ip = '127.0.' + str(i) + '.0'
  node = maindb.create_node(node_identifier, ip, 1234, '0.1a', True, 'the owner pubkey', extra_vessel_name)
  node_list.append(node)


# Create Donation records
donation_list = []
for i in range(100):
  node = node_list[i]
  donor = user_list[i % len(user_list)]
  donation = maindb.create_donation(node, donor, 'some resource description text')
  donation_list.append(donation)


# Create Vessel records
vessel_list = []
for i in range(100):
  node = node_list[i]
  name = 'v' + str(i)
  vessel = maindb.create_vessel(node, name)
  vessel_list.append(vessel)
  # Set the vessel ports.
  ports = range(1000, 1010)
  ports.append(user_list[0].usable_vessel_port)
  maindb.set_vessel_ports(vessel, ports)
  




from django.conf.urls.defaults import *

# Uncomment the next two lines to enable the admin:
# from django.contrib import admin
# admin.autodiscover()

urlpatterns = patterns('seattlegeni.website.html.views',
                       
                       # 'Get Donations' functions:
                       # show the main download page for downloading installers
                       (r'^(?P<username>\w{3,32})/$', 'download', {}, 'installers'),
                       # build and download the android installer
                       (r'^(?P<username>\w{3,32})/seattle_android.zip$', 'build_android_installer', {}, 'android_installer'),
                       # build and download the windows installer
                       (r'^(?P<username>\w{3,32})/seattle_win.zip$', 'build_win_installer', {}, 'win_installer'),
                       # build and download the linux installer
                       (r'^(?P<username>\w{3,32})/seattle_linux.tgz$', 'build_linux_installer', {}, 'linux_installer'),
                       # build and download the mac installer
                       (r'^(?P<username>\w{3,32})/seattle_mac.tgz$', 'build_mac_installer', {}, 'mac_installer'),
                       # help page
                       (r'^(?P<username>\w{3,32})/help$', 'donations_help', {}, 'donations_help'),
                      )



"""
<Program>
  mocklib.py
  
<Author>
  Justin Samuel

<Date Started>
  Aug 10, 2009
  
<Purpose>
  This provides functions that take care of mocking out (read: monkey
  patching) various parts of seattlegeni's internal api, including calls to:

    nodemanager.get_node_info
    lockserver.create_lockserver_handle
    lockserver.destroy_lockserver_handle
    lockserver._perform_lock_request
    backend.acquire_vessel
    backend.generate_key
    keygen.generate_keypair
    
  The idea is to use the functions provided in this module to keep test scripts
  clean and not repeat code in each one doing the same or similar monkey
  patching with mock versions of these function calls.
"""

from seattlegeni.common.api import backend
from seattlegeni.common.api import keygen
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import nodemanager

from seattlegeni.common.exceptions import *



_mock_nodemanager_get_node_info_args = None

def _mock_get_node_info(ip, port):
  
  (nodeid_key, version, vessels_dict) = _mock_nodemanager_get_node_info_args
  nodeinfo = {"version" : version,
              "nodename" : "",
              "nodekey" : nodeid_key,
              "vessels" : {}}
  nodeinfo["vessels"] = vessels_dict
  return nodeinfo

def mock_nodemanager_get_node_info(nodeid_key, version, vessels_dict):
  global _mock_nodemanager_get_node_info_args
  _mock_nodemanager_get_node_info_args = (nodeid_key, version, vessels_dict)
  nodemanager.get_node_info = _mock_get_node_info






def _mock_create_lockserver_handle(lockserver_url=None):
  pass

def _mock_destroy_lockserver_handle(lockserver_handle):
  pass

def _mock_perform_lock_request(request_type, lockserver_handle, user_list=None, node_list=None):
  pass

def mock_lockserver_calls():
  
  lockserver.create_lockserver_handle = _mock_create_lockserver_handle
  lockserver.destroy_lockserver_handle = _mock_destroy_lockserver_handle
  lockserver._perform_lock_request = _mock_perform_lock_request





_mock_acquire_vessel_result_list = None

def _mock_acquire_vessel(geniuser, vessel):
  
  result_list = _mock_acquire_vessel_result_list
  
  if len(result_list) == 0:
    raise Exception("_mock_acquire_vessel ran out results. " + 
                    "Either you need to provide more results in the result_list, " + 
                    "or this is a legitimate test failure.")
  if result_list.pop(0):
    pass
  else:
    raise UnableToAcquireResourcesError

def mock_backend_acquire_vessel(result_list):
  """
  Provide a list of boolean values that the mock'd out backend.acquire_vessels
  will use to decide whether to return without an exception (True) or to raise
  an UnableToAcquireResourcesError (False). The list will be used in order and
  an exception will be raised if the mock backend.acquire_vessel() function is
  called more times than there are items in the list.
  """
  global _mock_acquire_vessel_result_list
  _mock_acquire_vessel_result_list = result_list
  backend.acquire_vessel = _mock_acquire_vessel





_mock_generate_key_keylist = None

def _mock_generate_key(keydescription):
  keylist = _mock_generate_key_keylist
  
  if len(keylist) == 0:
    raise Exception("_mock_generate_key ran out of keys. " + 
                    "Either you need to provide more keys, or this is a legitimate test failure.")
  return keylist.pop(0)

def mock_backend_generate_key(keylist):
  global _mock_generate_key_keylist
  _mock_generate_key_keylist = keylist
  backend.generate_key = _mock_generate_key





_mock_generate_keypair_keypairlist = None

def _mock_generate_keypair():
  keypairlist = _mock_generate_keypair_keypairlist
  
  if len(keypairlist) == 0:
    raise Exception("_mock_generate_keypair ran out of keypairs. " + 
                    "Either you need to provide more keypairs, or this is a legitimate test failure.")
  return keypairlist.pop(0)

def mock_keygen_generate_keypair(keypairlist):
  global _mock_generate_keypair_keypairlist
  _mock_generate_keypair_keypairlist = keypairlist
  keygen.generate_keypair = _mock_generate_keypair




"""
<Program>
  errorviews.py

<Started>
  August 4, 2008

<Author>
  Justin Samuel

<Purpose>
  This module defines functions used for rendering error pages. We don't use the
  default error views because we want our 500 error to have access to the
  MEDIA_URL setting and therefore look like the rest of the site.
  
  We don't define a handler for 404 errors because the default one passes a
  RequestContext already, so we don't need to customize it to be able to
  access the MEDIA_URL value in templates. The 404.html file will be used
  by the default 404handler.
"""


from django.template import loader, RequestContext
from django import http



def internal_error(request):
  """
  The view that is used for rendering internal errors.
  This view function is referenced in the 500handler variable in urls.conf.
  """
  template = loader.get_template('500.html')
  return http.HttpResponseServerError(template.render(RequestContext(request, {})))



"""
<Program>
  dispatcher.py

<Started>
  6 July 2009

<Author>
  Justin Samuel

<Purpose>
  XMLRPC handler for django.

  Nothing in this file needs to be modified when adding/removing/changing
  public xmlrpc methods. For that, see the views.py file in the same directory.

  This xmlrpc dispatcher for django is modified from the version at:
  http://code.djangoproject.com/wiki/XML-RPC
"""

from SimpleXMLRPCServer import SimpleXMLRPCDispatcher
from django.http import HttpResponse

from seattlegeni.website.xmlrpc.views import PublicXMLRPCFunctions
from django.views.decorators.csrf import csrf_exempt


# This is the url that will be displayed if the xmlrpc service is requested
# directory through a web browser (that is, through a GET request).
SEATTLECLEARINGHOUSE_XMLRPC_API_DOC_URL = "https://seattle.cs.washington.edu/wiki/SeattleGeniApi"

# Create a Dispatcher. This handles the calls and translates info to function maps.
# TODO: allow_none = True or False? Does using None in the api make the xmlrpc
#       api python-specific?
dispatcher = SimpleXMLRPCDispatcher(allow_none=False, encoding=None)


@csrf_exempt
def rpc_handler(request):
  """
  All xmlrpc requests are initially routed here by django. The actual functions we
  implement are in the views.py file. This rpc_handler function will make sure that
  what we return from the functions in views.py will be turned into a valid xmlrpc
  response.
  
  If POST data is defined, it assumes it's XML-RPC and tries to process as such.
  If the POST data is empty or if it is a GET request, this assumes the request is
  from a browser and responds saying it's an xmlrpc service.
  
  This function does not need to be called from anywhere other than having it
  defined in urls.py. That is, you generally shouldn't need to ever use this
  function directly.
  """
  response = HttpResponse()
  if len(request.POST):
    response.write(dispatcher._marshaled_dispatch(request.raw_post_data))
  else:
    response.write("<b>This is the SeattleGeni XML-RPC Service.</b><br>")
    response.write("Please see <a href=" + SEATTLECLEARINGHOUSE_XMLRPC_API_DOC_URL + ">" + SEATTLECLEARINGHOUSE_XMLRPC_API_DOC_URL + "</a> for more information.")

  response['Content-length'] = str(len(response.content))
  return response



# All methods in the PublicXMLRPCFunctions class will be available as xmlrpc functions.
dispatcher.register_instance(PublicXMLRPCFunctions())



"""
<Program>
  views.py

<Started>
  September 4, 2009

<Author>
  Justin Samuel

<Purpose>
  Views for reports about seattlegeni.
"""

import django.contrib.auth.decorators
from django.http import HttpResponse

from seattlegeni.common.util import statistics






def user_is_staff_member(user):
  return user.is_authenticated() and user.is_staff





@django.contrib.auth.decorators.user_passes_test(user_is_staff_member)
def index(request):
  html = '<html><head><style>body { font-family : sans-serif; }</style></head><body>'
  html += '<h2>Available reports</h2>'
  html += 'Note: some of these reports are not generated efficiently. The ones '
  html += 'that take a long time to load are probably hammering the database.<br /><br />'
  html += '<ul>'
  html += '<li><a href="%s">%s</a></li>' % ("all", "All reports on one page") 
  html += '<li><a href="%s">%s</a></li>' % ("donations", "Donations")
  html += '<li><a href="%s">%s</a></li>' % ("acquired_vessels", "Acquired vessels")
  html += '<li><a href="%s">%s</a></li>' % ("vessels_by_port", "Available vessels by port")
  html += '<li><a href="%s">%s</a></li>' % ("lan_sizes_by_port", "Available LAN sizes by port")
  html += '</ul>'
  html += '</body></html>'
  return HttpResponse(html)





@django.contrib.auth.decorators.user_passes_test(user_is_staff_member)
def all(request):
  lines = []
  lines += _get_text_number_of_vessels_acquired_per_user()
  lines.append("")
  lines += _get_text_donations_per_user()
  lines.append("")
  lines += _get_text_available_vessels_by_port()
  lines.append("")
  lines += _get_text_lan_sizes_by_port()
  output = "\n".join(lines)
  return HttpResponse(output, content_type='text/plain')
  
  
  


@django.contrib.auth.decorators.user_passes_test(user_is_staff_member)
def acquired_vessels(request):
  output = "\n".join(_get_text_number_of_vessels_acquired_per_user())
  return HttpResponse(output, content_type='text/plain')
  
  



@django.contrib.auth.decorators.user_passes_test(user_is_staff_member)
def donations(request):
  output = "\n".join(_get_text_donations_per_user())
  return HttpResponse(output, content_type='text/plain')


  


@django.contrib.auth.decorators.user_passes_test(user_is_staff_member)
def vessels_by_port(request):
  output = "\n".join(_get_text_available_vessels_by_port())
  return HttpResponse(output, content_type='text/plain')





def lan_sizes_by_port(request):
  output = "\n".join(_get_text_lan_sizes_by_port())
  return HttpResponse(output, content_type='text/plain')

  



def _get_text_number_of_vessels_acquired_per_user():
  
  lines = []
  lines.append("Number of vessels acquired per user")
  lines.append("-----------------------------------")
  
  vessel_acquisition_dict = statistics.get_vessel_acquisition_counts_by_user()
  
  for username in vessel_acquisition_dict:
    lines.append(username + (" " * (25 - len(username))) + str(vessel_acquisition_dict[username]))
  
  lines.append("")
  lines.append("Vessel acquisition summary")
  lines.append("--------------------------")
  lines.append("There are " +  str(len(vessel_acquisition_dict.keys())) + " users who have acquired vessels.")
  lines.append("There are " + str(sum(vessel_acquisition_dict.values())) + " acquired vessels.")

  return lines





def _get_text_donations_per_user():
  
  lines = []
  lines.append("Donations per user")
  lines.append("------------------")
  
  donations_dict = statistics.get_donation_counts_by_user()
  
  lines.append("User" + (" " * (25 - len("User"))) + "Active" + "\t" + "Inactive")
  lines.append("----" + (" " * (25 - len("----"))) + "------" + "\t" + "--------")
    
  total_active_donations = 0
  total_inactive_donations = 0
    
  for username in donations_dict:
    (active_donation_count, inactive_donation_count) = donations_dict[username]
    total_active_donations += active_donation_count 
    total_inactive_donations +=  inactive_donation_count
    lines.append(username + (" " * (25 - len(username))) + str(active_donation_count) + "\t" + str(inactive_donation_count))

  lines.append("")
  lines.append("Donation summary")
  lines.append("--------------------------")
  lines.append("Total active donations:   " +  str(total_active_donations))
  lines.append("Total inactive donations: " +  str(total_inactive_donations))
  
  return lines

  

  
  
def _get_text_available_vessels_by_port():
  
  lines = []
  lines.append("Available vessels by port")
  lines.append("-------------------------")
  
  available_vessels_dict = statistics.get_available_vessel_counts_by_port()
  
  lines.append("%s\t%s\t%s\t%s" % ("Port", "All", "No-NAT", "Only-NAT"))
  lines.append("%s\t%s\t%s\t%s" % ("----", "---", "------", "--------"))
  
  for port in available_vessels_dict:
    portdict = available_vessels_dict[port]
    lines.append("%s\t%s\t%s\t%s" % (port, portdict["all"], portdict["no_nat"], portdict["only_nat"]))

  return lines





def _get_text_lan_sizes_by_port():
  
  lines = []
  lines.append("Available LAN sizes by port")
  lines.append("---------------------------")
  
  lan_sizes_by_port = statistics.get_available_lan_vessel_counts()
  
  lines.append("%s\t%s" % ("Port", "Size List (top five)"))
  lines.append("%s\t%s" % ("----", "--------------------"))
  
  for port in lan_sizes_by_port:
    size_list = lan_sizes_by_port[port][:5]
    lines.append("%s\t%s" % (port, size_list))

  return lines



"""
Currently just replaces the backend.acquire_vessel function to make it claim
to fail half of the time. 
"""

import random

from seattlegeni.common.api import backend

from seattlegeni.common.exceptions import UnableToAcquireResourcesError

print "Mocking: seattlegeni.common.api.backend.acquire_vessel"

def mock_acquire_vessel(geniuser, vessel):
  if random.randint(0, 1) == 1:
    raise UnableToAcquireResourcesError

backend.acquire_vessel = mock_acquire_vessel



import unittest

import lockserver_daemon as lockserver


class TheTestCase(unittest.TestCase):

  def setUp(self):
    # Reset the lockserver's global variables between each test.
    lockserver.init_globals()


  def testUserAndNodeLockContention_one(self):
    # Start three sessions.
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())

    # First session requests lock on the user 'bob'.
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[0], locks)
    
    # First session requests locks on the nodes '123' and '456'.
    locks = {'node':['123','456']}
    lockserver.do_acquire_locks(sess[0], locks)

    # Second session requests lock on the node '123'.
    locks = {'node':['123']}
    lockserver.do_acquire_locks(sess[1], locks)

    # Third session requests lock on the user 'bob'.
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[2], locks)

    expected_heldlockdict = {
      'user': {'bob': {'locked_by_session': sess[0],
                       'queue': [sess[2]]}},
      'node': {'123': {'locked_by_session': sess[0],
                       'queue': [sess[1]]},
               '456': {'locked_by_session': sess[0],
                       'queue': []}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': ['bob'], 'node': ['123','456']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False},
      sess[2]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': ['bob'], 'node': []},
                'acquirelocksproceedeventset': False}}

    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])

    # First session releases user lock on 'bob'.
    # Note: they still hold the node locks on '123' and '456'
    locks = {'user':['bob']}
    lockserver.do_release_locks(sess[0], locks)

    expected_heldlockdict = {
      'user': {'bob': {'locked_by_session': sess[2],
                       'queue': []}},
      'node': {'123': {'locked_by_session': sess[0],
                       'queue': [sess[1]]},
               '456': {'locked_by_session': sess[0],
                       'queue': []}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': [], 'node': ['123','456']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False},
      sess[2]: {'heldlocks': {'user': ['bob'], 'node': []},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True}}

    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])

    # First session releases node lock on '123'.
    # Note: they still hold the node lock on '456'
    locks = {'node':['123']}
    lockserver.do_release_locks(sess[0], locks)

    expected_heldlockdict = {
      'user': {'bob': {'locked_by_session': sess[2],
                       'queue': []}},
      'node': {'123': {'locked_by_session': sess[1],
                       'queue': []},
               '456': {'locked_by_session': sess[0],
                       'queue': []}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': [], 'node': ['456']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'user': [], 'node': ['123']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'user': ['bob'], 'node': []},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True}}

    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])

    # First session releases node lock on '456' and then requests locks on
    # user 'bob' and nodes '123' and '456' again. It can't request the node
    # locks, however, until it gets the node lock (in xmlrpc usage, the
    # user lock request would have blocked). So, we have the session that is
    # holding the lock on user 'bob' release that lock before the first session
    # makes the node lock requests.
    locks = {'node':['456']}
    lockserver.do_release_locks(sess[0], locks)
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[0], locks)
    locks = {'user':['bob']}
    lockserver.do_release_locks(sess[2], locks)
    locks = {'node':['123', '456']}
    lockserver.do_acquire_locks(sess[0], locks)
        
    expected_heldlockdict = {
      'user': {'bob': {'locked_by_session': sess[0],
                       'queue': []}},
      'node': {'123': {'locked_by_session': sess[1],
                       'queue': [sess[0]]},
               '456': {'locked_by_session': sess[0],
                       'queue': []}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': ['bob'], 'node': ['456']},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False},
      sess[1]: {'heldlocks': {'user': [], 'node': ['123']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True}}

    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])



"""
<Program>
  maindb.py

<Started>
  29 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is the API that should be used to interact with the Main Database.
  Functions in this module are the only way that other code should interact
  with the Main Database.
   
  Note: it is yet to be determined how strict the rule of "the only way to
  interact with the Main Database" should be. It is likely that functions
  in this module will return values that will, for example, lazily load
  data from the database. The point is to hide as much of the underlying
  database interaction from other code. It is definitely the case that
  no code outside this module should be modifying the database.
  
  We try to keep all manual transaction management in seattlegeni to within
  this module. The general idea is that the default behavior of django is
  what we want in most places (to commit any time data-altering functions
  are called, such as .save() or .delete()). However, in a few cases we
  want multiple data-altering functions to be committed atomically, so we
  use @transaction.commit_manually.
"""

# It is a bit confusing to just import datetime because then you have to use
# things like 'datetime.datetime.now()'. So, import the parts we need.
from datetime import datetime
from datetime import timedelta

import django.contrib.auth.models
import django.core.exceptions
import django.db

from django.db import transaction

import random

from seattlegeni.common.exceptions import *

from seattlegeni.common.util import log

from seattlegeni.common.util.decorators import log_function_call
from seattlegeni.common.util.decorators import log_function_call_and_only_first_argument

from seattlegeni.common.util.assertions import *

from seattlegeni.website import settings

from seattlegeni.website.control.models import Donation
from seattlegeni.website.control.models import GeniUser
from seattlegeni.website.control.models import Node
from seattlegeni.website.control.models import Vessel
from seattlegeni.website.control.models import VesselPort
from seattlegeni.website.control.models import VesselUserAccessMap
from seattlegeni.website.control.models import ActionLogEvent
from seattlegeni.website.control.models import ActionLogVesselDetails





# The number of free vessel credits each user gets regardless of donations.
# This is the value set for newly-created users. Changing this value only
# affects new user accounts, not existing ones.
DEFAULT_FREE_VESSEL_CREDITS = 10

# A user gets 10 vessel credits for every donation they make.
VESSEL_CREDITS_FOR_DONATIONS_MULTIPLIER = 10

# Port from this range will be randomly selected as the usable_vessel_port for
# new users. Remember that range(x, y) is x inclusive, y exclusive. 
ALLOWED_USER_PORTS = range(63100, 63180)

# Number of ascii characters in generated API keys.
API_KEY_LENGTH = 32

# When initially acquired, this is the amount of time from now until a vessel
# will expire.
# This must be a datetime.timedelta object.
DEFAULT_VESSEL_EXPIRATION_TIMEDELTA = timedelta(hours=4)

# When a vessel is renewed, this is the amount of time from now until a vessel
# will expire.
MAXIMUM_VESSEL_EXPIRATION_TIMEDELTA = timedelta(days=7)

# When calls to get_available_*_vessels() are made, we try to return more
# vessels than were requested so that there will be extras in case some
# can't be acquired. The number we try to return is the number requested times
# the multiplier plus the adder. This gives the chance for half of the nodes
# to be unavailable but the adder makes sure small numbers of requested
# vessels (e.g. 1) won't fail due to really bad luck with the returned vessels.
GET_AVAILABLE_VESSELS_MULTIPLIER = 2
GET_AVAILABLE_VESSELS_ADDER = 10

# The maximum number of lists of same-subnet vessels to return from a call to
# get_available_lan_vessels_by_subnet(). This is useful because if, for example,
# the user requested one lan vessel, we might return (not to the user) a list that
# includes a list of vessels for every possible subnet covered by active nodes.
# That wouldn't be the end of the world, but it seems a bit excessive if we have
# nodes on thousands of subnets.
GET_AVAILABLE_LAN_VESSELS_MAX_SUBNETS = 10

# The string that is the prefix to all NAT strings in node last_known_ip fields.
NAT_STRING_PREFIX = "NAT$"





@log_function_call
def init_maindb():
  """
  <Purpose>
    Initializes the database in a way that makes database transaction commits
    from other sources immediately visible. Must be called after creating
    any database connection.
    
    If you're using the maindb in long-running non-website code, you should
    either call this function on a regular basis (even if new database
    connections haven't been made) or copy the bit of code here that does the
    django.db.reset_queries() call to your own code. Otherwise, your memory
    usage will grow due to query logging when DEBUG is True.
  <Arguments>
    None.
  <Exceptions>
    It is possible for a Django / Database error to be thrown (which is fatal).
  <Side Effects>
    This function makes it so that we see any transactions commits made by
    other database clients even within a single transaction on the side
    of the script that calls this function.
    This is done by changing the transaction isolation level of InnoDB from the
    default of "repeatable read" to instead be "read committed". For more info, see:
    http://dev.mysql.com/doc/refman/5.4/en/innodb-consistent-read.html
    http://dev.mysql.com/doc/refman/5.4/en/set-transaction.html
  <Returns>
    None.
  """
  if settings.DATABASE_ENGINE is "mysql":
    django.db.connection.cursor().execute('set transaction isolation level read committed')
  else:
    log.error("init_maindb() called when not using mysql. This is only OK when developing.")

  # We shouldn't be running in production with settings.DEBUG = True. Just in
  # case, though, tell django to reset its list of saved queries. On the
  # website, init_maindb() will get called with each web request so we'll be
  # resetting the queries at the beginning of each request.
  # See http://docs.djangoproject.com/en/dev/faq/models/#why-is-django-leaking-memory
  # for more info.
  if settings.DEBUG:
    log.debug("Resetting django query log because settings.DEBUG is True.")
    log.error("Reminder: settings.DEBUG is True. Don't run in production like this!")
    django.db.reset_queries()



@transaction.commit_manually
@log_function_call_and_only_first_argument
def create_user(username, password, email, affiliation, user_pubkey, user_privkey, donor_pubkey):
  """
  <Purpose>
    Create a new seattlegeni user in the database. This user will be able to
    login to website, use the seattlegeni xmlrpc api, acquire vessels, etc.
    
    A 'user' lock should be held on the specified username before calling this
    function. The code calling this function should have already checked to
    see whether a user by this username exists (and did so while holding the
    lock on that username).
  
    The code calling this function is responsible for first storing the
    corresponding private key of donor_pubkey in the keydb. In the case of the
    website calling this function, the website will make a call to the backend
    to request a new key and that will take care of storing the donor private
    key in the keydb.
    
  <Arguments>
    username
      The username of the user to be created.
    password
      The user's password.
    email
      The user's email address.
    affiliation
      The affiliation text provided by the user.
    user_pubkey
      A string of the user's public key.
    user_privkey
      A string of the user's private key or None. This would be None if the
      user has provided their own public key rather than had us generate
      a keypair for them.
    donor_pubkey
      A string of the user's donor key. This is a key the user never sees (and
      probably never knows exists).  
      
    <Exceptions>
      None
      
    <Side Effects>
      Creates a user record in the django user table and a user record in the
      seattlegeni geniuser table, the two of which have a one-to-one mapping.
      Does not change the database if creation of either record fails.
      
    <Returns>
      A GeniUser object of the newly created user.
  """
  assert_str(username)
  assert_str(password)
  assert_str(email)
  assert_str(affiliation)
  assert_str(user_pubkey)
  assert_str_or_none(user_privkey)
  assert_str(donor_pubkey)
  
  # We're committing manually to make sure the multiple database writes are
  # atomic. (That is, regenerate_api_key() will do a database write.)
  try:
    # Generate a random port for the user's usable vessel port.
    port = random.sample(ALLOWED_USER_PORTS, 1)[0]
  
    # Create the GeniUser (this is actually records in two different tables
    # underneath because of model inheretance, but django hides that from us).
    geniuser = GeniUser(username=username, password='', email=email,
                        affiliation=affiliation, user_pubkey=user_pubkey,
                        user_privkey=user_privkey, donor_pubkey=donor_pubkey,
                        usable_vessel_port=port,
                        free_vessel_credits=DEFAULT_FREE_VESSEL_CREDITS)
    # Set the password using this function so that it gets hashed by django.
    geniuser.set_password(password)
    geniuser.save()
  
    regenerate_api_key(geniuser)
    
  except:
    transaction.rollback()
    raise
  
  else:
    transaction.commit()

  return geniuser





@log_function_call
def regenerate_api_key(geniuser):
  """
  <Purpose>
    Set a new, randomly generated api key for a user.
  <Arguments>
    geniuser
      The GeniUser object of the user whose api key is to be regenerated.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with a new,
    randomly-generated api key.
  <Returns>
    The new api key that has been set for the user.
  """
  assert_geniuser(geniuser)
  
  # Create a set of potential characters for the api key that includes the
  # numbers 0-9 and the uppercase charactesr A-Z, excluding the letter 'O'
  # because it looks too much like zero.
  population = []
  population.extend(range(ord('0'), ord('9') + 1))
  population.extend(range(ord('A'), ord('Z') + 1))
  population.remove(ord('O'))
  
  api_key = ""
  for character in random.sample(population, API_KEY_LENGTH):
    api_key += chr(character)
    
  geniuser.api_key = api_key
  geniuser.save()

  return api_key





@log_function_call
def set_user_keys(geniuser, pubkey, privkey):
  """
  <Purpose>
    Sets the public/private user keys for the geniuser.
  <Arguments>
    geniuser
      The GeniUser object of the user whose keys are to be changed.
    pubkey
      The public key string of the public key to be set for the user.
    privkey
      The private key string of the private key to be set for the user. This
      can be None if no private key should be stored in the database.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with the
    provided keys.
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  
  geniuser.user_pubkey = pubkey
  geniuser.user_privkey = privkey
  geniuser.save()





@log_function_call
def set_user_email(geniuser, new_email):
  """
  <Purpose>
    Sets the email for the geniuser.
  <Arguments>
    geniuser
      The GeniUser object of the user whose email is to be changed.
    new_email  
      The new email value.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with the
    provided email.
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  
  geniuser.email = new_email
  geniuser.save()





@log_function_call
def set_user_affiliation(geniuser, new_affiliation):
  """
  <Purpose>
    Sets the affiliation for the geniuser.
  <Arguments>
    geniuser
      The GeniUser object of the user whose affiliation is to be changed.
    new_affiliation  
      The new affiliation value.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with the
    provided affiliation.
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  
  geniuser.affiliation = new_affiliation
  geniuser.save()





@log_function_call
def set_user_port(geniuser, new_port): # currently not in use 
  """
  <Purpose>
    Sets the port for the geniuser.
  <Arguments>
    geniuser
      The GeniUser object of the user whose port is to be changed.
    new_port  
      The new port value.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with the
    provided port.
  <Returns>
    None
  """
  # if new_port in ALLOWED_USER_PORTS:
  assert_geniuser(geniuser)
  
  geniuser.usable_vessel_port = new_port
  geniuser.save()





@log_function_call_and_only_first_argument
def set_user_password(geniuser, new_password):
  """
  <Purpose>
    Sets the password for the geniuser.
  <Arguments>
    geniuser
      The GeniUser object of the user whose password is to be changed.
    new_password  
      The new password value.
  <Exceptions>
    None
  <Side Effects>
    Updates the database as well as the geniuser object passed in with the
    provided password.
  <Returns>
    None
  """
  assert_geniuser(geniuser)

  # Set the password using this function so that it gets hashed by django.
  geniuser.set_password(new_password)
  geniuser.save()





@log_function_call
def create_node(node_identifier, last_known_ip, last_known_port, last_known_version, is_active, owner_pubkey, extra_vessel_name):
  """
  <Purpose>
    Create a new node record in the database. A node lock should be held before
    calling this function.
  <Arguments>
    node_identifier
      The identifier of the node to be created (there must be any existing
      nodes with this identifier).
    last_known_ip
      The last known ip address (a string) that this node's nodemanager was
      running on.
    last_known_port
      The last known port (an int) that this node's nodemanager was running on.
    last_known_version
      The last known version of Seattle (a string) that this node was running.
    is_active
      Whether this node is considered to be up.
    owner_pubkey
      The owner public key (a string) that SeattleGeni uses for this node. The
      corresponding private key must be stored in the keydb.
    extra_vessel_name
      The name of the 'extra vessel' on this node (a string).
  <Exceptions>
    None
  <Side Effects>
    A node record is created in the database.
  <Returns>
    The Node object of the created node.
  """
  assert_str(node_identifier)
  assert_str(last_known_ip)
  assert_int(last_known_port)
  assert_str(last_known_version)
  assert_bool(is_active)
  assert_str(owner_pubkey)
  assert_str(extra_vessel_name)
  
  # Make sure there is not already a node with this node identifier.
  try:
    get_node(node_identifier)
    raise ProgrammerError("A node with this identifier already exists: " + node_identifier)
  except DoesNotExistError:
    pass
  
  # Create the Node.
  node = Node(node_identifier=node_identifier, last_known_ip=last_known_ip,
              last_known_port=last_known_port, last_known_version=last_known_version,
              is_active=is_active, owner_pubkey=owner_pubkey,
              extra_vessel_name=extra_vessel_name, is_broken=False)
  node.save()

  return node





def get_allowed_user_ports():
  """
  <Purpose>
    Gets the allowed user ports defined globally in this file.
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    The allowed user ports which are a list.
  """
  return ALLOWED_USER_PORTS





@log_function_call
def create_donation(node, donor, resource_description_text):
  """
  <Purpose>
    Create a new donation record in the database. A node lock and a user lock
    should be held before calling this function.
  <Arguments>
    node
      The Node object of the node that the donation was made from.
    donor
      The GeniUser object of the user that made the donation.
    resource_description_text
      A description of the donated resources (in the format of other resource
      descriptions used in Seattle).
  <Exceptions>
    None
  <Side Effects>
    A donation record is created in the database.
  <Returns>
    The Donation object of the created donation.
  """
  assert_node(node)
  assert_geniuser(donor)
  assert_str(resource_description_text)
  
  # Create the Donation.
  donation = Donation(node=node, donor=donor,
                      resource_description_text=resource_description_text)
  donation.save()

  return donation





@log_function_call
def create_vessel(node, vesselname):
  """
  <Purpose>
    Create a new vessel record in the database. A node lock should be held
    before calling this function.
  <Arguments>
    node
      The Node object of the node that the vessel exists on.
    vesselname
      The name of the vessel for the vessel record to be created.
  <Exceptions>
    None
  <Side Effects>
    A vessel record is created in the database.
  <Returns>
    The Vessel object of the created vessel.
  """
  assert_node(node)
  assert_str(vesselname)
  
  # Create the Vessel.
  vessel = Vessel(node=node, name=vesselname, acquired_by_user=None,
                  date_acquired=None, date_expires=None, is_dirty=False,
                  user_keys_in_sync=True)
  vessel.save()
  
  return vessel





@transaction.commit_manually
@log_function_call
def set_vessel_ports(vessel, port_list):
  """
  <Purpose>
    Change the list of ports the database considers associated with a vessel.
    A node lock should be held before calling this function.
  <Arguments>
    vessel
      The Vessel object whose list of ports is to be changed.
    port_list
      The list of port numbers (int's or long's) that are the complete list
      of ports for this vessel.
  <Exceptions>
    None
  <Side Effects>
    The database indicates that the ports in port_list (and only those ports)
    are the ports for the vessel. 
  <Returns>
    None
  """
  assert_vessel(vessel)
  assert_list(port_list)
  for port in port_list:
    assert_int(port)

  # We're committing manually to make sure the multiple database writes are
  # atomic.
  try:
    # Delete all existing VesselPort records for this vessel.
    VesselPort.objects.filter(vessel=vessel).delete()
    
    # Create a VesselPort record for each port in port_list.
    for port in port_list:
      vesselport = VesselPort(vessel=vessel, port=port)
      vesselport.save()

  except:
    transaction.rollback()
    raise
  
  else:
    transaction.commit()





@log_function_call
def get_users_with_access_to_vessel(vessel):
  """
  <Purpose>
    Determine which users have access to a vessel according to the database.
  <Arguments>
    vessel
      The Vessel object whose user access is info is wanted.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of GeniUser objects of the users who have access to the vessel.
  """
  assert_vessel(vessel)

  user_list = []

  for vmap in VesselUserAccessMap.objects.filter(vessel=vessel):
    user_list.append(vmap.user)

  return user_list





@log_function_call
def get_vessels_accessible_by_user(geniuser):
  """
  <Purpose>
    Determine which vessels the database indicates the user has access to.
  <Arguments>
    geniuser
      The GeniUser object of the user whose vessel access info is wanted.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Vessel objects of the vessels the user has access to.
  """
  assert_geniuser(geniuser)
  
  vessel_list = []
  
  for vmap in VesselUserAccessMap.objects.filter(user=geniuser):
    vessel_list.append(vmap.vessel)

  return vessel_list





@log_function_call
def add_vessel_access_user(vessel, geniuser):
  """
  <Purpose>
    Indicate in the database that a user has access to a vessel. A node lock
    and a user lock should be held before calling this function.
  <Arguments>
    vessel
      The Vessel object that the user is being given access to.
    geniuser
      The GeniUser object of the user that being access to the vessel.
  <Exceptions>
    None
  <Side Effects>
    The database indicates that the user has access to the vessel.
  <Returns>
    None
  """
  assert_vessel(vessel)
  assert_geniuser(geniuser)

  # If the user already has access, don't throw an exception, just consider
  # the request done.
  mapqueryset = VesselUserAccessMap.objects.filter(vessel=vessel, user=geniuser)
  if mapqueryset.count() == 1:
    return
  
  # Create a VesselUserAccessMap record.
  maprecord = VesselUserAccessMap(vessel=vessel, user=geniuser)
  maprecord.save()
  
  
  
  
  
@log_function_call
def remove_vessel_access_user(vessel, geniuser):
  """
  <Purpose>
    Indicate in the database that a user no longer has access to a vessel. A
    node lock and a user lock should be held before calling this function.
  <Arguments>
    vessel
      The Vessel object that the user is having access removed from.
    geniuser
      The GeniUser object of the user that is having access removed.
  <Exceptions>
    None
  <Side Effects>
    The database no longer indicates that the user has access to the vessel.
  <Returns>
    None
  """
  assert_vessel(vessel)
  assert_geniuser(geniuser)

  # Delete any map records for this user/vessel.
  VesselUserAccessMap.objects.filter(vessel=vessel, user=geniuser).delete()





@log_function_call
def _remove_all_user_access_to_vessel(vessel):
  """
  <Purpose>
    Indicate in the database that no user has access to the vessel.
  <Arguments>
    vessel
      The Vessel object to have all user access removed.
  <Exceptions>
    None
  <Side Effects>
    The database no longer indicates that any user has access to the vessel.
  <Returns>
    None
  """
  assert_vessel(vessel)

  # Delete any map records for this user/vessel.
  VesselUserAccessMap.objects.filter(vessel=vessel).delete()





@log_function_call
def get_user(username, allow_inactive=False):
  """
  <Purpose>
    Retrieve the user that a has the given username.
  <Arguments>
    username
      The username of the user to be retrieved.
    allow_inactive
      If True, don't raise an exception if the account is inactive (which
      means probably banned for being a bad citizen).
  <Exceptions>
    DoesNotExistError
      If there is no user with the given username.
  <Side Effects>
    None
  <Returns>
    The GeniUser object of the user.
  """
  assert_str(username)
  
  try:
    geniuser = GeniUser.objects.get(username=username)
    if not geniuser.is_active and not allow_inactive:
      # They've probably been banned. We could completely lie, but if this is
      # really a banned account then they probably know they've been banned.
      # If it's not someone who was up to no good, then they should be
      # encouraged to contact us (at least, if they get to see this message).
      raise DoesNotExistError("Account must be activated. Please contact support.")
    
  except django.core.exceptions.ObjectDoesNotExist:
    # Intentionally vague message to prevent a security problem if this ever
    # gets displayed on the frontend (needs to be the same message as the
    # password/apikey authentication failure messages).
    raise DoesNotExistError("No such user.")
    
  except django.core.exceptions.MultipleObjectsReturned:
    raise InternalError("Multiple records returned when looking up a user by username.")

  return geniuser





@log_function_call_and_only_first_argument
def get_user_with_password(username, password):
  """
  <Purpose>
    Retrieve the user that a has the given username and password.
  <Arguments>
    username
      The username of the user to be retrieved.
    password
      The password of the user to be retrieved.
  <Exceptions>
    DoesNotExistError
      If there is no user with the given username and password.
  <Side Effects>
    None
  <Returns>
    The GeniUser object of the user.
  """
  assert_str(username)
  assert_str(password)
  
  # Throws a DoesNotExistError if there is no such user.
  geniuser = get_user(username)

  if not django.contrib.auth.models.check_password(password, geniuser.password):
    # Intentionally vague message to prevent a security problem if this ever
    # gets displayed on the frontend.
    raise DoesNotExistError("No such user.")
  
  return geniuser





@log_function_call_and_only_first_argument
def get_user_with_api_key(username, api_key):
  """
  <Purpose>
    Retrieve the user that a has the given username and api_key.
  <Arguments>
    username
      The username of the user to be retrieved.
    api_key
      The api_key of the user to be retrieved.
  <Exceptions>
    DoesNotExistError
      If there is no user with the given username and api_key.
  <Side Effects>
    None
  <Returns>
    The GeniUser object of the user.
  """
  assert_str(username)
  assert_str(api_key)
  
  # Throws a DoesNotExistError if there is no such user.
  geniuser = get_user(username)

  if not geniuser.api_key == api_key:
    # Intentionally vague message to prevent a security problem if this ever
    # gets displayed on the frontend.
    raise DoesNotExistError("No such user.")
  
  return geniuser




@log_function_call
def get_donor(donor_pubkey):
  """
  <Purpose>
    Retrieve the user that has the donor_pubkey,
  <Arguments>
    donor_pubkey
      The key that is the user's donor key.
  <Exceptions>
    DoesNotExistError
      If there is no user with the given donor_pubkey.
  <Side Effects>
    None
  <Returns>
    The GeniUser object of the user who is the donor.
  """
  assert_str(donor_pubkey)
  
  try:
    geniuser = GeniUser.objects.get(donor_pubkey=donor_pubkey)
    
  except django.core.exceptions.ObjectDoesNotExist:
    raise DoesNotExistError("No user exists with the specified donor_pubkey.")
    
  except django.core.exceptions.MultipleObjectsReturned:
    raise InternalError("Multiple records returned when looking up a user by donor_pubkey.")

  return geniuser





@log_function_call
def get_donations_from_node(node):
  """
  <Purpose>
    Retrieve the donation records of the donations from a node.
  <Arguments>
    node
      The Node object of the node from which the donations were made.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Donation objects.
  """
  assert_node(node)
  
  return list(Donation.objects.filter(node=node))





@log_function_call
def get_node_identifier_from_vessel(vessel):
  """
  <Purpose>
    Determine the node id of the node a vessel is on.
  <Arguments>
    vessel
      The Vessel object of the vessel whose node's nodeid will be retrieved.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    The node id (a string).
  """
  assert_vessel(vessel)
  
  return vessel.node.node_identifier





@log_function_call
def delete_user_private_key(geniuser):
  """
  <Purpose>
    Delete the user's private user key. A user lock should be held before
    calling this function.
  <Arguments>
    geniuser
      The user whose private key is to be deleted.
  <Exceptions>
    None
  <Side Effects>
    The user's private key has been removed from the database.
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  
  geniuser.user_privkey = None
  geniuser.save()





@log_function_call
def get_donations_by_user(geniuser, include_inactive_and_broken=False):
  """
  <Purpose>
    Retrieve a list of all donations made by a user. By default, only includes
    donations from nodes that are active and not broken.
  <Arguments>
    geniuser
      The user whose donations are to be retrieved.
    include_inactive_and_broken
      Whether to include donations by the user that are from nodes which are
      inactive and/or broken. Default is False.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Donation objects.
  """
  assert_geniuser(geniuser)
  
  queryset = Donation.objects.filter(donor=geniuser)
  if not include_inactive_and_broken:
    queryset = queryset.filter(node__is_active=True)
    queryset = queryset.filter(node__is_broken=False)
  
  # Let's return it as a list() rather than a django QuerySet.
  # Using list() causes the QuerySet to be converted to a list, which also
  # means the query is executed (no lazy loading).
  return list(queryset)





@log_function_call
def get_node(node_identifier):
  """
  <Purpose>
    Retrieve a Node object that represents a specific node.
  <Arguments>
    node_identifier
      The identifier of the node to be retrieved.
  <Exceptions>
    DoesNotExistError
      If there is no node in the database with the provided identifier.
  <Side Effects>
    None
  <Returns>
    A Node object.
  """
  assert_str(node_identifier)

  try:
    node = Node.objects.get(node_identifier=node_identifier)
    
  except django.core.exceptions.ObjectDoesNotExist:
    raise DoesNotExistError("There is no node with the node identifier: " + str(node_identifier))
    
  except django.core.exceptions.MultipleObjectsReturned:
    raise InternalError("Multiple records returned when looking up a node by node identifier.")
  
  return node





@log_function_call
def set_node_owner_pubkey(node, ownerkeystring):
  """
  <Purpose>
    Change a node's owner key. A node lock should be held before calling this
    function.
  <Arguments>
    node
      The node object of the node whose owner key is to be modified.
    ownerkeystring
      The public key string to be set as the node's owner key.
  <Exceptions>
    None
  <Side Effects>
    The node's owner key is changed in the database and the node object passed
    in to the function reflects this.
  <Returns>
    None
  """
  assert_node(node)
  assert_str(ownerkeystring)
  
  node.owner_pubkey = ownerkeystring
  node.save()





@log_function_call
def record_node_communication_failure(node):
  """
  <Purpose>
    Let the database know that communication with a node has failed.
  <Arguments>
    node
      The node object of the node that couldn't be communicated with.
  <Exceptions>
    None
  <Side Effects>
    The node's is_active value is changed to False in the database and the
    node object passed in to the function reflects this.
  <Returns>
    None
  """
  assert_node(node)
  
  node.is_active = False
  node.save()





@log_function_call
def record_node_communication_success(node, version, ip, port):
  """
  <Purpose>
    Let the database know that communication with a node has succeeded and
    provide updated information about the node. This is really only intended
    to be used by transition scripts.
  <Arguments>
    node
      The Node object of the node that was successfully communicated with.
    version
      The version of seattle the node reports it is running.
    ip
      The ip address or NAT string the nodemanager is currently accessible at.
    port
      The port the nodemanager is currently accessible at.
  <Exceptions>
    None
  <Side Effects>
    The node's is_active value is changed to True in the database if it
    wasn't already, the date_last_contacted value is updated in the database
    to the current time, and the version, ip and port are updated. The node
    object passed in to the function reflects these changes.
  <Returns>
    None
  """
  assert_node(node)
  assert_str(version)
  assert_str(ip)
  assert_positive_int(port)
  
  node.last_known_version = version
  node.last_known_ip = ip
  node.last_known_port = port
  node.is_active = True
  node.date_last_contacted = datetime.now()
  node.save()





@log_function_call
def set_node_extra_vessel_name(node, extra_vessel_name):
  """
  <Purpose>
    Let the database know the name of the extra vessel on a node has changed.
    This is only intended to be used by transition scripts.
  <Arguments>
    node
      The Node object of the node that was successfully communicated with.
    extra_vessel_name
      The new name of the extra vessel on the node.
  <Exceptions>
    None
  <Side Effects>
    The node's extra_vessel_name value is changed in the database and the node
    object passed in to the function reflects the change.
  <Returns>
    None
  """
  assert_node(node)
  assert_str(extra_vessel_name)
  
  node.extra_vessel_name = extra_vessel_name
  node.save()





@log_function_call
def mark_node_as_active(node):
  """
  <Purpose>
    Let the database know that the node is active. This should only be used
    when a node moves from the canonical state to onepercent state. The
    onepercent-to-onepercent transition script will use the other method
    called record_node_communication_success(), instead.
  <Arguments>
    node
      The Node object of the node that is active.
  <Exceptions>
    None
  <Side Effects>
    The node's is_active value is changed to True in the database if it
    wasn't already. The node object passed to the function is correspondingly
    updated.
  <Returns>
    None
  """
  assert_node(node)
  
  node.is_active = True
  node.save()





@log_function_call
def mark_node_as_inactive(node):
  """
  <Purpose>
    Let the database know that the node is inactive. You shouldn't use this
    method if communication with a node fails. Instead, only use this if there
    is some other reason to mark the node as inactive (e.g. the node is up
    but the nodeid is wrong, indicating it's not seattle instance we expected).
  <Arguments>
    node
      The Node object of the node that is broken.
  <Exceptions>
    None
  <Side Effects>
    The node's is_broken value is changed to True in the database if it
    wasn't already. The node object passed to the function is correspondingly
    updated.
  <Returns>
    None
  """
  assert_node(node)
  
  node.is_active = False
  node.save()





@log_function_call
def mark_node_as_broken(node):
  """
  <Purpose>
    Let the database know that the node is broken (that is, what's one the node
    doesn't match our database). There currently isn't a corresponding function
    to mark the node as not broken because fixing a node will be a manual
    process.
  <Arguments>
    node
      The Node object of the node that is broken.
  <Exceptions>
    None
  <Side Effects>
    The node's is_broken value is changed to True in the database if it
    wasn't already. The node object passed to the function is correspondingly
    updated.
  <Returns>
    None
  """
  assert_node(node)
  
  node.is_broken = True
  node.save()





@log_function_call
def get_vessel(node_identifier, vesselname):
  """
  <Purpose>
    Retrieve a Vessel object that represents a specific vessel.
  <Arguments>
    node_identifier
      The identifier of the node that the vessel is on.
    vesselname
      The name of the vessel.
  <Exceptions>
    DoesNotExistError
      If there is no such vessel in the database (including if there is no node
      with the given identifier).
  <Side Effects>
    None
  <Returns>
    A Vessel object.
  """
  assert_str(node_identifier)
  assert_str(vesselname)

  try:
    vessel = Vessel.objects.get(node__node_identifier=node_identifier, name=vesselname)
    
  except django.core.exceptions.ObjectDoesNotExist:
    raise DoesNotExistError("There is no vessel with the node identifier: " + 
                            str(node_identifier) + " and vessel name: " + vesselname)
    
  except django.core.exceptions.MultipleObjectsReturned:
    raise InternalError("Multiple records returned when looking up a vessel by node identifier and vessel name.")
  
  return vessel





@log_function_call
def get_acquired_vessels(geniuser):
  """
  <Purpose>
    Retrieve a list of vessels that are acquired by a user.
  <Arguments>
    geniuser
      The GeniUser object of the user whose acquired vessels are to be
      retrieved.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Vessel objects.
  """
  assert_geniuser(geniuser)
  
  # Let's return it as a list() rather than a django QuerySet.
  # Using list() causes the QuerySet to be converted to a list, which also
  # means the query is executed (no lazy loading).
  queryset = Vessel.objects.filter(acquired_by_user=geniuser)
  # We don't include expired vessels. That is, as far as what the user sees and
  # is considered having acquired, vessels expire immediately when their
  # expiration time arrives.
  queryset = queryset.exclude(date_expires__lte=datetime.now())
  return list(queryset)
  




def _get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(port):
  """
  Get a queryset of vessels that have a certain port and which are all
  available to be acquired. There is no restriction on whether these are
  vessels on nat nodes.
  """
  queryset = Vessel.objects.filter(acquired_by_user=None)
  # No dirty vessels or inactive nodes.
  queryset = queryset.exclude(is_dirty=True)
  queryset = queryset.exclude(node__is_active=False)
  queryset = queryset.exclude(node__is_broken=True)
  # Make sure we only get vessels with the user's assigned port.
  queryset = queryset.filter(vesselport__port__exact=port)
  # Randomize the vessels returned by the query.
  # Using order_by('?') is the QuerySet way of saying ORDER BY RAND().
  queryset = queryset.order_by('?')
  
  log.debug("There are " + str(queryset.count()) + " available NAT and non-NAT node vessels on port " + str(port))

  return queryset





def _get_queryset_of_all_available_vessels_for_a_port_exclude_nat_nodes(port):
  """
  Get a queryset of vessels that have a certain port and which are all
  available to be acquired. Excludes vessels on nat nodes.
  """
  queryset = _get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(port)
  
  queryset = queryset.exclude(node__last_known_ip__startswith=NAT_STRING_PREFIX)
  
  log.debug("There are " + str(queryset.count()) + " available non-NAT node vessels on port " + str(port))

  return queryset





def _get_queryset_of_all_available_vessels_for_a_port_only_nat_nodes(port):
  """
  Get a queryset of vessels that have a certain port and which are all
  available to be acquired. Only includes vessels on nat nodes.
  """
  queryset = _get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(port)
  
  queryset = queryset.filter(node__last_known_ip__startswith=NAT_STRING_PREFIX)
  
  log.debug("There are " + str(queryset.count()) + " available NAT node vessels on port " + str(port))

  return queryset





@log_function_call
def get_available_rand_vessels(geniuser, vesselcount):
  """
  <Purpose>
    Get a list of potential vessels that could satisfy the user's request for
    rand vessels.
  <Arguments>
    geniuser
      The user who is requesting vessels.
    veselcount
      The number of rand vessels the user has requested.
  <Exceptions>
     UnableToAcquireResourcesError
       If there are not enough resources in seattlegeni to fulfill the request.
  <Side Effects>
    None
  <Returns>
    A list of vessels that contains at least vesselcount vessels. These may be
    any type of vessels (some could be on nat nodes, some could be on the same
    subnet, etc.).
  """
  assert_geniuser(geniuser)
  assert_positive_int(vesselcount)
  
  # We return more vessels than were asked for. This gives some room for some
  # of the vessels to be inaccessible or already acquired by the time they are
  # attempted to be acquired by the client code.
  returnvesselcount = GET_AVAILABLE_VESSELS_MULTIPLIER * vesselcount + GET_AVAILABLE_VESSELS_ADDER
  
  allvesselsqueryset = _get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(geniuser.usable_vessel_port)
   
  if allvesselsqueryset.count() < vesselcount:
    message = "Requested " + str(vesselcount) + " rand vessels, but we only have " + str(allvesselsqueryset.count())
    message += " vessels with port " + str(geniuser.usable_vessel_port) + " available." 
    raise UnableToAcquireResourcesError(message)
  
  return list(allvesselsqueryset[:returnvesselcount])





@log_function_call
def get_available_nat_vessels(geniuser, vesselcount):
  """
  <Purpose>
    Get a list of potential vessels that could satisfy the user's request for
    nat vessels.
  <Arguments>
    geniuser
      The user who is requesting vessels.
    veselcount
      The number of nat vessels the user has requested.
  <Exceptions>
     UnableToAcquireResourcesError
       If there are not enough resources in seattlegeni to fulfill the request.
  <Side Effects>
    None
  <Returns>
    A list of vessels that contains at least vesselcount vessels. All vessels
    in the list will be nat vessels.
  """
  assert_geniuser(geniuser)
  assert_positive_int(vesselcount)
  
  # We return more vessels than were asked for. This gives some room for some
  # of the vessels to be inaccessible or already acquired by the time they are
  # attempted to be acquired by the client code.
  returnvesselcount = GET_AVAILABLE_VESSELS_MULTIPLIER * vesselcount + GET_AVAILABLE_VESSELS_ADDER
  
  natvesselsqueryset = _get_queryset_of_all_available_vessels_for_a_port_only_nat_nodes(geniuser.usable_vessel_port)
   
  if natvesselsqueryset.count() < vesselcount:
    #COMMENT this out when NAT acquistion feature is re-enabled  /* ADDED Aug 06, 2012 by GP */
    message = 'Acquiring NAT vessels is currently disabled. '
    #UNCOMMENT this when NAT acquistion feature is re-enabled
    #message = "Requested " + str(vesselcount) + " nat vessels, but we only have " + str(natvesselsqueryset.count())
    #message += " vessels with port " + str(geniuser.usable_vessel_port) + " available." 
    raise UnableToAcquireResourcesError(message)
  
  return list(natvesselsqueryset[:returnvesselcount])





@log_function_call
def get_available_wan_vessels(geniuser, vesselcount):
  """
  <Purpose>
    Get a list of potential vessels that could satisfy the user's request for
    wan vessels.
  <Arguments>
    geniuser
      The user who is requesting vessels.
    veselcount
      The number of wan vessels the user has requested.
  <Exceptions>
     UnableToAcquireResourcesError
       If there are not enough resources in seattlegeni to fulfill the request.
  <Side Effects>
    None
  <Returns>
    A list of vessels that contains at least vesselcount vessels. No two
    vessels in the list will have the same last octet of their IP address.
  """
  assert_geniuser(geniuser)
  assert_positive_int(vesselcount)
  
  # We return more vessels than were asked for. This gives some room for some
  # of the vessels to be inaccessible or already acquired by the time they are
  # attempted to be acquired by the client code.
  returnvesselcount = GET_AVAILABLE_VESSELS_MULTIPLIER * vesselcount + GET_AVAILABLE_VESSELS_ADDER
  
  vessellist = []
  includedsubnets = []
  
  nonnatvesselsqueryset = _get_queryset_of_all_available_vessels_for_a_port_exclude_nat_nodes(geniuser.usable_vessel_port)
   
  # Note: it would be more efficient to have the sql query return vessels
  # in unique subnets, but we would have to be very careful that the UNIQUE
  # wasn't being applied/interpreted "before" the RAND(). That is, we would
  # have to be careful that we weren't always getting the same node for a given
  # subnet. So, instead of worrying about the sql for that which wouldn't
  # be intuitive to do with the django ORM, let's just do that part manually.
  
  for possiblevessel in nonnatvesselsqueryset:
    
    subnet = possiblevessel.node.last_known_ip.rpartition('.')[0]
    if not subnet:
      log.error("The vessel " + str(possiblevessel) + " has an invalid last_known_ip")
      continue
    
    # For efficiency, includedsubnets should be a constant time lookup type.
    if subnet in includedsubnets:
      continue
    
    includedsubnets.append(subnet)
    vessellist.append(possiblevessel)
    
    if len(vessellist) == returnvesselcount:
      break 

  if len(vessellist) < vesselcount:
    message = "Requested " + str(vesselcount) + " wan vessels, but we only have vessels with port "
    message += str(geniuser.usable_vessel_port) + " available on " + str(len(includedsubnets)) + " subnets." 
    raise UnableToAcquireResourcesError(message)
  
  return vessellist





def _get_subnet_list():
  """
  Returns a randomly-ordered list of subnets that have at least one active
  non-nat node on the subnet.
  """
  
  # The commented out code only works on mysql but is more efficient. Leaving
  # it out for now as it's harder to test, as our tests run on sqlite.
  
  # Get a list of subnets that have at least vesselcount active nodes in the subnet.
  # This doesn't guarantee that there are available vessels for this user on
  # those nodes, but it's a start. We'll narrow it down further after we get
  # the list of possible subnets.
  #
  #  subnetsql = """SELECT COUNT(*) AS lansize,
  #                        SUBSTRING_INDEX(last_known_ip, '.', 3) AS subnet
  #                 FROM control_node
  #                 WHERE is_active = TRUE AND
  #                       is_broken = FALSE
  #                 GROUP BY subnet
  #                 HAVING lansize >= %s
  #                 ORDER BY RAND()""" % (vesselcount)
  #                 
  #  cursor = django.db.connection.cursor()
  #  cursor.execute(subnetsql)
  #  return cursor.fetchall()
  
  # Get a queryset of all active nodes.
  queryset = Node.objects.filter(is_active=True)
  queryset = queryset.filter(is_broken=False)
  queryset = queryset.exclude(last_known_ip__startswith=NAT_STRING_PREFIX)
  queryset = queryset.order_by('last_known_ip')

  previous_subnet = None
  subnetlist = []
  
  # Look through the queryset and build a list of unique subnets.
  for node in queryset:
    
    subnet = node.last_known_ip.rpartition('.')[0]
    
    if not subnet:
      log.error("The node " + str(node) + " has an invalid last_known_ip")
      continue
    
    if subnet != previous_subnet:
      subnetlist.append(subnet)
      previous_subnet = subnet
           
  # Randomize the order of the subnets.
  random.shuffle(subnetlist)
  
  return subnetlist
  




@log_function_call
def get_available_lan_vessels_by_subnet(geniuser, vesselcount):
  """
  <Purpose>
    Get a list of potential subnets (and a list of vessels in the subnet)
    that could satisfy the user's request for lan vessels.
  <Arguments>
    geniuser
      The user who is requesting vessels.
    veselcount
      The number of lan vessels the user has requested.
  <Exceptions>
     UnableToAcquireResourcesError
       If there are not enough resources in seattlegeni to fulfill the request.
  <Side Effects>
    None
  <Returns>
    A list of vessel lists. Each vessel list contains vessels on the same
    subnet. Each vessel list has a minimum of vesselcount availalable vessels
    on the user's port.
    
    The number of vessel lists (that is, the number of different subnets)
    returned may not include all possible subnets that seattlegeni knows about.
    This is for efficiency reasons. The maximum number of subnets returned
    can be adjusted through the GET_AVAILABLE_LAN_VESSELS_MAX_SUBNETS
    constant at the top of this module.
  """
  assert_geniuser(geniuser)
  assert_positive_int(vesselcount)
  
  subnetlist = _get_subnet_list()

  if len(subnetlist) == 0:
    raise UnableToAcquireResourcesError("No subnets exist with at least " + str(vesselcount) + " active nodes")
  
  subnets_vessels_list = []
  
  nonnatvesselsqueryset = _get_queryset_of_all_available_vessels_for_a_port_exclude_nat_nodes(geniuser.usable_vessel_port)
  
  for subnet in subnetlist:
    lanvesselsqueryset = nonnatvesselsqueryset.filter(node__last_known_ip__startswith=subnet + '.')
    
    if lanvesselsqueryset.count() >= vesselcount:
      # We don't worry about too many vessels being in this list, as it will be 255 at most.
      subnets_vessels_list.append(list(lanvesselsqueryset))

    if len(subnets_vessels_list) >= GET_AVAILABLE_LAN_VESSELS_MAX_SUBNETS:
      break

  if len(subnets_vessels_list) == 0:
    message = "No subnets exist with at least " + str(vesselcount)
    message += " active nodes that have a vessel available on port " + str(geniuser.usable_vessel_port)
    raise UnableToAcquireResourcesError(message)
  
  return subnets_vessels_list





@log_function_call
def get_user_free_vessel_credits(geniuser):
  """
  <Purpose>
    Determine number of free vessel credits the user gets (that is, vessel
    credits they get from registering an account without having donated any
    resources).
  <Arguments>
    geniuser
      The GeniUser object of the user whose freee vessel credit count is to be
      retrieved.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    The user's number of free vessel credits.
  """
  assert_geniuser(geniuser)

  return geniuser.free_vessel_credits





@log_function_call
def get_user_vessel_credits_from_donations(geniuser):
  """
  <Purpose>
    Determine number of vessel credits the user has earned due to donations.
  <Arguments>
    geniuser
      The GeniUser object of the user whose vessel credits from donations are to
      be retrieved.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    The user's number of vessel credits from donations.
  """
  assert_geniuser(geniuser)

  return len(get_donations_by_user(geniuser)) * VESSEL_CREDITS_FOR_DONATIONS_MULTIPLIER
  




@log_function_call
def get_user_total_vessel_credits(geniuser):
  """
  <Purpose>
    Determine the total number of vessel credits the user has, regardless of
    the number of vessels that already have acquired. This is the sum of the
    number of free vessel credits for the user and the number of vessel
    credits from donations by the user.
  <Arguments>
    geniuser
      The GeniUser object of the user whose total vessel credit count is to be
      retrieved.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    The user's total number of vessel credits.
  """
  return get_user_free_vessel_credits(geniuser) + get_user_vessel_credits_from_donations(geniuser) 
  




@log_function_call
def require_user_can_acquire_resources(geniuser, requested_vessel_count):
  """
  <Purpose>
    Ensure the user is allowed to acquire these resources. For now, only
    checks that a user has enough vessel credits for what they have requested.
    In the future, it could also check if the user was "banned".
    This isn't named assert_* because it doesn't raise an AssertionError
    when the user isn't allowed acquire resources.
  <Arguments>
    geniuser
    requested_vessel_count
  <Exceptions>
    InsufficientUserResourcesError
      If the user doesn't have enough vessel credits to satisfy the number
      of vessels the have requested.
  <Side Effects>
    None
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  assert_int(requested_vessel_count)
  
  # This isn't in the assertions module because it is all database-related
  # information. I wanted to avoid making the assertions module dependent
  # on maindb. If it's confusing, the name of this method can be changed.
  
  # This can be made faster. When necessary, add a get_acquired_vessels_count() function.
  acquired_vessel_count = len(get_acquired_vessels(geniuser))
  
  max_allowed_vessels = get_user_total_vessel_credits(geniuser)
  
  if requested_vessel_count + acquired_vessel_count > max_allowed_vessels:
    raise InsufficientUserResourcesError("Requested " + str(requested_vessel_count) + 
                                         " vessels, already acquired " + str(acquired_vessel_count) + 
                                         ", max allowed is " + str(max_allowed_vessels))





@log_function_call
def record_acquired_vessel(geniuser, vessel):
  """
  <Purpose>
    Performs all database operations necessary to record the fact that a vessel
    was acquired by a user.
  <Arguments>
    geniuser
      The GeniUser object of the user who is acquiring the vessel.
    vessel
      The Vessel object to be marked as acquired by the user.
  <Exceptions>
    None
  <Side Effects>
    The vessel is marked as acquired by geniuser. The user and vessel are
    also added to the database's VesselUserAccessMap table.
  <Returns>
    None
  """
  assert_geniuser(geniuser)
  assert_vessel(vessel)
  
  # We aren't caching any information with the user record about how many
  # resources have been acquired, so the only thing we need to do is make the
  # vessel as having been acquired by this user.
  vessel.acquired_by_user = geniuser
  vessel.date_acquired = datetime.now()
  vessel.date_expires = vessel.date_acquired + DEFAULT_VESSEL_EXPIRATION_TIMEDELTA
  vessel.save()
  
  # Update the database to reflect that this user has access to this vessel.
  add_vessel_access_user(vessel, geniuser)





@log_function_call
def record_released_vessel(vessel):
  """
  <Purpose>
    Performs all database operations necessary to record the fact that a vessel
    was released or expired.
  <Arguments>
    vessel
      The Vessel object to be marked as having been released.
  <Exceptions>
    None
  <Side Effects>
    The vessel is marked as not acquired by any user as well as marked as dirty.
    All records for this vessel in the database's VesselUserAccessMap have been
    removed.
  <Returns>
    None
  """
  assert_vessel(vessel)
  
  # We remove the VesselUserAccessMap records first just in case something
  # fails. That is, we don't want to leave the VesselUserAccessMap records
  # around with the vessel later getting cleaned up and given to another user.
  # Alternatively, we could manually commit a transaction for the
  # record_released_vessel() function that we're in to make sure all or none
  # of it gets done.
  _remove_all_user_access_to_vessel(vessel)
  
  # We aren't caching any information with the user record about how many
  # resources have been acquired, so the only thing we need to do is make the
  # vessel as having not having been acquired by any user.
  vessel.acquired_by_user = None
  vessel.is_dirty = True
  vessel.user_keys_in_sync = True
  vessel.date_acquired = None
  vessel.date_expires = None
  vessel.save()





def set_maximum_vessel_expiration(vessel):
  """
  <Purpose>
    Sets an acquired vessel's expiration date to the maximum allowed.
  <Arguments>
    vessel
      The Vessel object whose acquisition is to be extended.
  <Exceptions>
    None
  <Side Effects>
    The vessel's expiration date has been extended.
  <Returns>
    None
  """
  assert_vessel(vessel)
  
  vessel.date_acquired = datetime.now()
  vessel.date_expires = vessel.date_acquired + MAXIMUM_VESSEL_EXPIRATION_TIMEDELTA
  vessel.save()





# We don't log the function call here so that we don't fill up the backend
# daemon's logs.
def mark_expired_vessels_as_dirty():
  """
  <Purpose>
    Change all vessel records in the database whose acquisitions have expired
    to be marked as dirty in the database (that is, to indicate they need to
    be cleaned up by the backend).
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    All expired vessels (past expiration and acquired by users) in the database
    are marked as dirty as well as marked as not acquired by users. Additionally,
    all vessel user access map entries for each of these vessels have been
    removed.
  <Returns>
    The list of vessels that were marked as dirty.
  """
  # We want to mark as dirty all vessels past their expiration date that are
  # currently acquired by users.
  queryset = Vessel.objects.filter(date_expires__lte=datetime.now())
  queryset = queryset.exclude(acquired_by_user=None)
  
  vessel_list = list(queryset)
  
  if len(vessel_list) > 0:
    for vessel in vessel_list:
      record_released_vessel(vessel)

  # Return the number of vessels that just expired.
  return vessel_list





@log_function_call
def mark_vessel_as_clean(vessel):
  """
  <Purpose>
    Change the database to indicate that a vessel has been cleaned up by the
    backend.
  <Arguments>
    vessel
      The Vessel object of the vessel to be marked as clean.
  <Exceptions>
    None
  <Side Effects>
    Marks the vessel as clean in the database.
  <Returns>
    None
  """
  assert_vessel(vessel)
  
  vessel.is_dirty = False
  vessel.user_keys_in_sync = True
  vessel.save()





# We don't log the function call here so that we don't fill up the backend
# daemon's logs.
def get_vessels_needing_cleanup():
  """
  <Purpose>
    Determine which vessels need to be cleaned up by the backend.
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Vessel objects which are the vessels needing to be cleaned up.
  """
  queryset = Vessel.objects.filter(is_dirty=True)
  queryset = queryset.filter(node__is_active=True)
  queryset = queryset.filter(node__is_broken=False)
  # Be certain not to clean up vessels acquired by users. This is here mostly
  # in case an admin marked a vessel for cleanup that was acquired by a user.
  queryset = queryset.filter(acquired_by_user=None)
  return list(queryset)






@log_function_call
def does_vessel_need_cleanup(vessel):
  """
  <Purpose>
    Determine whether a given vessel needs to be cleaned up by the backend.
  <Arguments>
    vessel
      The Vessel object that we want to know if it needs to be cleaned up.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A tuple (needs_cleanup, reason) where needs_cleanup is a boolean indicating
    whether the vessel needs cleanup and reason is a string that indicates the
    reason cleanup is not needed if needs_cleanup is False.
  """
  assert_vessel(vessel)
  
  # Re-query the database in case this vessel has changed or been deleted.
  try:
    vessel = Vessel.objects.get(id=vessel.id)

  except django.core.exceptions.ObjectDoesNotExist:
    # The vessel was deleted.
    return (False, "The vessel no longer exists")
  
  if not vessel.node.is_active:
    return (False, "The node the vessel is on is not active")
  
  if vessel.node.is_broken:
    return (False, "The node the vessel is on is broken")
  
  if not vessel.is_dirty:
    return (False, "The vessel is not dirty")
  
  if vessel.acquired_by_user is not None:
    return (False, "The vessel is currently acquired")
  
  return (True, "")
  




def mark_vessel_as_needing_user_key_sync(vessel):
  """
  <Purpose>
    Set a vessel's database record to indicate that the vessel's user keys need
    to be sync'd.
  <Arguments>
    vessel
      The Vessel object of the vessel whose user keys are out of sync.
  <Exceptions>
    None
  <Side Effects>
    The database record the vessel is updated.
  <Returns>
    None.
  """
  assert_vessel(vessel)
  
  vessel.user_keys_in_sync = False
  vessel.save()





def mark_vessel_as_not_needing_user_key_sync(vessel):
  """
  <Purpose>
    Set a vessel's database record to indicate that the vessel's user keys do
    not need to be sync'd.
  <Arguments>
    vessel
      The Vessel object of the vessel whose user keys are in sync.
  <Exceptions>
    None
  <Side Effects>
    The database record the vessel is updated.
  <Returns>
    None.
  """
  assert_vessel(vessel)
  
  vessel.user_keys_in_sync = True
  vessel.save()





def get_vessels_needing_user_key_sync():
  """
  <Purpose>
    Determine which vessels need to have user keys sync'd by the backend.
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Vessel objects which are the vessels needing to have their
    user keys sync'd.
  """
  queryset = Vessel.objects.filter(user_keys_in_sync=False)
  queryset = queryset.filter(is_dirty=False)
  queryset = queryset.filter(node__is_active=True)
  queryset = queryset.filter(node__is_broken=False)
  
  return list(queryset)





def does_vessel_need_user_key_sync(vessel):
  """
  <Purpose>
    Determine whether a given vessel needs to have its user keys sync'd the
    backend.
  <Arguments>
    vessel
      The Vessel object that we want to know if it needs to have user keys
      sync'd.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A tuple (needs_sync, reason) where needs_sync is a boolean indicating
    whether the vessel needs user keys sync'd and reason is a string that
    indicates the reason key sync is not needed if needs_sync is False.
  """
  assert_vessel(vessel)
  
  # Re-query the database in case this vessel has changed or been deleted.
  try:
    vessel = Vessel.objects.get(id=vessel.id)

  except django.core.exceptions.ObjectDoesNotExist:
    # The vessel was deleted.
    return (False, "The vessel no longer exists")
  
  if not vessel.node.is_active:
    return (False, "The node the vessel is on is not active")
  
  if vessel.node.is_broken:
    return (False, "The node the vessel is on is broken")
  
  if vessel.is_dirty:
    return (False, "The vessel is dirty")
  
  return (True, "")





@log_function_call
def delete_all_vessels_of_node(node):
  """
  <Purpose>
    Delete from the database all vessel records of a node.
  <Arguments>
    node
      The Node object whose vessel records are to be deleted.
  <Exceptions>
    None
  <Side Effects>
    All vessel records for this node have been removed from the database.
  <Returns>
    None
  """
  assert_node(node)
  
  Vessel.objects.filter(node=node).delete()





def get_active_nodes():
  """
  <Purpose>
    Get a list of all active nodes that are not broken.
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Node objects of active nodes that aren't broken.
  """
  return list(Node.objects.filter(is_active=True, is_broken=False))





def get_active_nodes_include_broken():
  """
  <Purpose>
    Get a list of all active nodes including those that are broken.
  <Arguments>
    None
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Node objects of active nodes.
  """
  return list(Node.objects.filter(is_active=True))





def get_vessels_on_node(node):
  """
  <Purpose>
    Get a list of all vessels on a nodes.
  <Arguments>
    node
      The Node object of the node we are interested in.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    A list of Vessel objects of vessels on the node.
  """
  return list(node.vessel_set.all())




def create_action_log_event(function_name, user, second_arg, third_arg,
                            was_successful, message, date_started, vessel_list):
  
  delta = datetime.now() - date_started
  completion_time = ((delta.days * 3600 * 24) + 
                     delta.seconds + 
                     (delta.microseconds / 1000000.0))
  
  event = ActionLogEvent(function_name=function_name,
                         user=user,
                         second_arg=second_arg,
                         third_arg=third_arg,
                         was_successful=was_successful,
                         message=message,
                         vessel_count=len(vessel_list),
                         date_started=date_started,
                         completion_time=completion_time)
  
  event.save()
  
  for vessel in vessel_list:
    vesseldetails = ActionLogVesselDetails(event=event, node=vessel.node,
                                           node_address=vessel.node.last_known_ip,
                                           node_port=vessel.node.last_known_port,
                                           vessel_name=vessel.name)
    vesseldetails.save()



from distutils.core import setup
setup(
  name = 'clearinghouse',
  packages = ['clearinghouse'], # this must be the same as the name above
  version = '0.1',
  description = "Seattle's seash package",
  author = 'Seattle',
  author_email = 'cc4256@nyu.edu',
  url = 'https://github.com/choksi81/clearinghouse', # use the URL to the github repo
  download_url = 'https://github.com/choksi81/clearinghouse/tarball/0.1', # I'll explain this in a second
  classifiers = [],
)
